from __future__ import annotations

from inspect import iscoroutine
from json import loads
from pathlib import Path
from time import sleep
from typing import TYPE_CHECKING, Any, Generator, Iterator, cast

import msgspec
import pytest
from starlette.responses import HTMLResponse, JSONResponse, PlainTextResponse
from starlette.responses import Response as StarletteResponse

from litestar import HttpMethod, Litestar, MediaType, Request, Response, get, route
from litestar._signature import SignatureModel
from litestar.background_tasks import BackgroundTask
from litestar.contrib.jinja import JinjaTemplateEngine
from litestar.datastructures import Cookie, ResponseHeader
from litestar.response import ServerSentEvent
from litestar.response.base import ASGIResponse
from litestar.response.file import ASGIFileResponse, File
from litestar.response.redirect import Redirect
from litestar.response.streaming import ASGIStreamingResponse, Stream
from litestar.response.template import Template
from litestar.status_codes import HTTP_200_OK, HTTP_308_PERMANENT_REDIRECT
from litestar.template.config import TemplateConfig
from litestar.testing import RequestFactory, create_test_client
from litestar.types import StreamType
from litestar.utils import AsyncIteratorWrapper
from litestar.utils.signature import ParsedSignature
from tests.models import DataclassPerson, DataclassPersonFactory

if TYPE_CHECKING:
    from typing import AsyncGenerator

    from litestar.routes import HTTPRoute


def my_generator() -> Generator[str, None, None]:
    for count in range(1, 11):
        yield str(count)
    return


async def my_async_generator() -> AsyncGenerator[str, None]:
    for count in range(1, 11):
        yield str(count)
    return


class MySyncIterator:
    def __init__(self) -> None:
        self.delay = 0.01
        self.i = 1
        self.to = 10

    def __iter__(self) -> Iterator[str]:
        return self

    def __next__(self) -> str:
        i = self.i
        if i > self.to:
            raise StopIteration
        self.i += 1
        if i:
            sleep(self.delay)
        return str(i)


class MyAsyncIterator(AsyncIteratorWrapper[str]):
    def __init__(self) -> None:
        super().__init__(iterator=MySyncIterator())


async def test_to_response_async_await(anyio_backend: str) -> None:
    @route(http_method=HttpMethod.POST, path="/person")
    async def handler(data: DataclassPerson) -> DataclassPerson:
        assert isinstance(data, DataclassPerson)
        return data

    person_instance = DataclassPersonFactory.build()
    handler._signature_model = SignatureModel.create(
        dependency_name_set=set(),
        fn=handler.fn,
        data_dto=None,
        parsed_signature=ParsedSignature.from_fn(handler.fn, {}),
        type_decoders=[],
    )

    response = await handler.to_response(
        data=handler.fn(data=person_instance),
        app=Litestar(route_handlers=[handler]),
        request=RequestFactory().get(route_handler=handler),
    )
    assert loads(response.body) == msgspec.to_builtins(person_instance)  # type: ignore[attr-defined]


async def test_to_response_returning_litestar_response() -> None:
    @get(path="/test")
    def handler() -> Response:
        return Response(media_type=MediaType.TEXT, content="ok")

    with create_test_client(handler) as client:
        http_route: HTTPRoute = client.app.routes[0]
        route_handler = http_route.route_handlers[0]
        response = await route_handler.to_response(
            data=route_handler.fn(), app=client.app, request=RequestFactory().get()
        )
        assert isinstance(response, ASGIResponse)


@pytest.mark.parametrize(
    "expected_response",
    [
        StarletteResponse(status_code=HTTP_200_OK, content=b"abc"),
        PlainTextResponse(content="abc"),
        HTMLResponse(content="<div><span/></div"),
        JSONResponse(status_code=HTTP_200_OK, content={}),
    ],
)
async def test_to_response_returning_starlette_response(
    expected_response: StarletteResponse, anyio_backend: str
) -> None:
    @get(path="/test")
    def handler() -> StarletteResponse:
        return expected_response

    with create_test_client(handler) as client:
        http_route: HTTPRoute = client.app.routes[0]
        route_handler = http_route.route_handlers[0]
        response = await route_handler.to_response(
            data=route_handler.fn(), app=client.app, request=RequestFactory().get()
        )
        assert isinstance(response, StarletteResponse)
        assert response is expected_response  # type: ignore[unreachable]


async def test_to_response_returning_redirect_response(anyio_backend: str) -> None:
    background_task = BackgroundTask(lambda: "")

    @get(
        path="/test",
        status_code=301,
        response_headers=[ResponseHeader(name="local-header", value="123")],
        response_cookies=[Cookie(key="redirect-cookie", value="aaa"), Cookie(key="general-cookie", value="xxx")],
    )
    def handler() -> Redirect:
        return Redirect(
            path="/somewhere-else",
            headers={"response-header": "abc"},
            cookies=[Cookie(key="redirect-cookie", value="xyz")],
            background=background_task,
        )

    with create_test_client(handler) as client:
        route: HTTPRoute = client.app.routes[0]
        route_handler = route.route_handlers[0]
        response = await route_handler.to_response(
            data=route_handler.fn(), app=client.app, request=RequestFactory().get()
        )
        encoded_headers = response.encode_headers()  # type: ignore[attr-defined]

        assert isinstance(response, ASGIResponse)
        assert (b"location", b"/somewhere-else") in encoded_headers
        assert (b"local-header", b"123") in encoded_headers
        assert (b"response-header", b"abc") in encoded_headers
        assert (b"set-cookie", b"general-cookie=xxx; Path=/; SameSite=lax") in encoded_headers
        assert (b"set-cookie", b"redirect-cookie=xyz; Path=/; SameSite=lax") in encoded_headers
        assert response.background == background_task


def test_to_response_returning_redirect_response_from_redirect() -> None:
    @get(path="/proxy")
    def proxy_handler() -> dict:
        return {"message": "redirected by before request hook"}

    def before_request_hook_handler(_: Request) -> Redirect:
        return Redirect(path="/proxy", status_code=HTTP_308_PERMANENT_REDIRECT)

    @get(path="/test", before_request=before_request_hook_handler)
    def redirect_handler() -> None:
        raise AssertionError("this endpoint should not be reached")

    with create_test_client(route_handlers=[redirect_handler, proxy_handler]) as client:
        response = client.get("/test")
        assert response.status_code == HTTP_200_OK
        assert response.json() == {"message": "redirected by before request hook"}


async def test_to_response_returning_file_response(anyio_backend: str) -> None:
    current_file_path = Path(__file__).resolve()
    filename = Path(__file__).name
    background_task = BackgroundTask(lambda: "")

    @get(
        path="/test",
        response_headers=[ResponseHeader(name="local-header", value="123")],
        response_cookies=[Cookie(key="redirect-cookie", value="aaa"), Cookie(key="general-cookie", value="xxx")],
    )
    def handler() -> File:
        return File(
            path=current_file_path,
            filename=filename,
            headers={"response-header": "abc"},
            cookies=[Cookie(key="file-cookie", value="xyz")],
            background=background_task,
        )

    with create_test_client(handler) as client:
        route: HTTPRoute = client.app.routes[0]
        route_handler = route.route_handlers[0]
        response = await route_handler.to_response(
            data=route_handler.fn(), app=client.app, request=RequestFactory().get()
        )
        assert isinstance(response, ASGIFileResponse)
        assert response.file_info
        if iscoroutine(response.file_info):
            await response.file_info

        encoded_headers = response.encode_headers()
        assert (b"local-header", b"123") in encoded_headers
        assert (b"response-header", b"abc") in encoded_headers
        assert (b"set-cookie", b"file-cookie=xyz; Path=/; SameSite=lax") in encoded_headers
        assert (b"set-cookie", b"general-cookie=xxx; Path=/; SameSite=lax") in encoded_headers
        assert (b"set-cookie", b"redirect-cookie=aaa; Path=/; SameSite=lax") in encoded_headers
        assert response.background == background_task


@pytest.mark.parametrize(
    "iterator, should_raise",
    [
        [my_generator(), False],
        [my_async_generator(), False],
        [MySyncIterator(), False],
        [MyAsyncIterator(), False],
        [my_generator, False],
        [my_async_generator, False],
        [MyAsyncIterator, False],
        [MySyncIterator, False],
        [[1, 2, 3, 4], False],
        ["abc", False],
        [b"abc", False],
        [{"key": 1}, False],
        [[{"key": 1}], False],
        [1, True],
        [None, True],
    ],
)
async def test_to_response_streaming_response(iterator: Any, should_raise: bool, anyio_backend: str) -> None:
    background_task = BackgroundTask(lambda: "")

    @get(
        path="/test",
        response_headers=[ResponseHeader(name="local-header", value="123")],
        response_cookies=[Cookie(key="redirect-cookie", value="aaa"), Cookie(key="general-cookie", value="xxx")],
    )
    def handler() -> Stream:
        return Stream(
            iterator,
            headers={"response-header": "abc"},
            cookies=[Cookie(key="streaming-cookie", value="xyz")],
            background=background_task,
        )

    with create_test_client(handler) as client:
        route: HTTPRoute = client.app.routes[0]
        route_handler = route.route_handlers[0]
        if should_raise:
            with pytest.raises(TypeError):
                await route_handler.to_response(data=route_handler.fn(), app=client.app, request=RequestFactory().get())
        else:
            response = await route_handler.to_response(
                data=route_handler.fn(), app=client.app, request=RequestFactory().get()
            )
            assert isinstance(response, ASGIStreamingResponse)
            encoded_headers = response.encode_headers()
            assert (b"local-header", b"123") in encoded_headers
            assert (b"response-header", b"abc") in encoded_headers
            assert (b"set-cookie", b"general-cookie=xxx; Path=/; SameSite=lax") in encoded_headers
            assert (b"set-cookie", b"redirect-cookie=aaa; Path=/; SameSite=lax") in encoded_headers
            assert (b"set-cookie", b"streaming-cookie=xyz; Path=/; SameSite=lax") in encoded_headers
            assert response.background == background_task


async def test_to_response_template_response(anyio_backend: str, tmp_path: Path) -> None:
    background_task = BackgroundTask(lambda: "")

    p = tmp_path / "test.template"
    p.write_text("<h1>hello world</h1>")

    @get(
        path="/test",
        response_headers=[ResponseHeader(name="local-header", value="123")],
        response_cookies=[Cookie(key="redirect-cookie", value="aaa"), Cookie(key="general-cookie", value="xxx")],
    )
    def handler() -> Template:
        return Template(
            template_name="test.template",
            context={},
            headers={"response-header": "abc"},
            cookies=[Cookie(key="template-cookie", value="xyz")],
            background=background_task,
        )

    app = Litestar(
        route_handlers=[],
        template_config=TemplateConfig(
            engine=JinjaTemplateEngine,
            directory=tmp_path,
        ),
    )

    with create_test_client(
        handler, template_config=TemplateConfig(engine=JinjaTemplateEngine, directory=tmp_path)
    ) as client:
        route: HTTPRoute = client.app.routes[0]
        route_handler = route.route_handlers[0]
        response = await route_handler.to_response(
            data=route_handler.fn(), app=client.app, request=RequestFactory(app=app).get()
        )
        assert isinstance(response, ASGIResponse)
        encoded_headers = response.encode_headers()

        assert (b"local-header", b"123") in encoded_headers
        assert (b"response-header", b"abc") in encoded_headers
        assert (b"set-cookie", b"general-cookie=xxx; Path=/; SameSite=lax") in encoded_headers
        assert (b"set-cookie", b"template-cookie=xyz; Path=/; SameSite=lax") in encoded_headers
        assert response.background == background_task


@pytest.mark.parametrize(
    "content",
    [
        my_generator(),
        my_async_generator(),
        MySyncIterator(),
        MyAsyncIterator(),
        my_generator,
        my_async_generator,
        MyAsyncIterator,
        MySyncIterator,
        [1, 2, 3, 4],
        "abc",
        b"abc",
        {"key": 1},
        [{"key": 1}],
    ],
)
async def test_to_response_sse_events(content: str | bytes | StreamType[str | bytes]) -> None:
    background_task = BackgroundTask(lambda: "")

    @get(
        path="/test",
    )
    def handler() -> ServerSentEvent:
        return ServerSentEvent(
            content=content,
            headers={"response-header": "abc"},
            cookies=[Cookie(key="streaming-cookie", value="xyz")],
            background=background_task,
            comment_message="my comment message\r\nwith some\nmixed line breaks",
            event_id=123,
            event_type="special",
        )

    with create_test_client(handler) as client:
        route: HTTPRoute = client.app.routes[0]
        route_handler = route.route_handlers[0]
        response = await route_handler.to_response(
            data=route_handler.fn(), app=client.app, request=RequestFactory().get()
        )
        encoded_headers = response.encode_headers()  # type: ignore[attr-defined]

        assert isinstance(response, ASGIStreamingResponse)
        assert ((b"cache-control", b"no-cache")) in encoded_headers
        assert (b"x-accel-buffering", b"no") in encoded_headers
        assert (b"connection", b"keep-alive") in encoded_headers
        assert (b"content-type", b"text/event-stream; charset=utf-8") in encoded_headers
        assert (b"response-header", b"abc") in encoded_headers
        assert (b"set-cookie", b"streaming-cookie=xyz; Path=/; SameSite=lax") in encoded_headers
        assert response.background == background_task


@pytest.mark.parametrize(
    "content",
    [
        my_generator(),
        my_async_generator(),
        my_generator,
        my_async_generator,
        MySyncIterator(),
        MyAsyncIterator(),
        MyAsyncIterator,
        MySyncIterator,
    ],
)
async def test_sse_events_content(content: str | bytes | StreamType[str | bytes]) -> None:
    @get(
        path="/test",
    )
    def handler() -> ServerSentEvent:
        return ServerSentEvent(
            content=content,
            comment_message="my comment message\r\nwith some\nmixed line breaks",
            event_id=123,
            event_type="special",
        )

    events: list[bytes] = []

    with create_test_client(handler) as client:
        route: HTTPRoute = client.app.routes[0]
        route_handler = route.route_handlers[0]
        response = await route_handler.to_response(
            data=route_handler.fn(), app=client.app, request=RequestFactory().get()
        )
        assert isinstance(response, ASGIStreamingResponse)
        async for value in response.iterator:
            events.append(cast("bytes", value))

    assert events == [
        b": my comment message\r\n",
        b": with some\r\n",
        b": mixed line breaks\r\n",
        b"id: 123\r\n",
        b"event: special\r\n",
        b"data: 1\r\n",
        b"data: 2\r\n",
        b"data: 3\r\n",
        b"data: 4\r\n",
        b"data: 5\r\n",
        b"data: 6\r\n",
        b"data: 7\r\n",
        b"data: 8\r\n",
        b"data: 9\r\n",
        b"data: 10\r\n",
        b"\r\n",
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, TypedDict, Union

__all__ = (
    "HtmxHeaderType",
    "LocationType",
    "TriggerEventType",
)

if TYPE_CHECKING:
    from typing_extensions import Required


EventAfterType = Literal["receive", "settle", "swap", None]

PushUrlType = Union[str, bool]

ReSwapMethod = Literal[
    "innerHTML", "outerHTML", "beforebegin", "afterbegin", "beforeend", "afterend", "delete", "none", None
]


class LocationType(TypedDict):
    """Type for HX-Location header."""

    path: Required[str]
    source: str | None
    event: str | None
    target: str | None
    swap: ReSwapMethod | None
    values: dict[str, str] | None
    hx_headers: dict[str, Any] | None


class TriggerEventType(TypedDict):
    """Type for HX-Trigger header."""

    name: Required[str]
    params: dict[str, Any] | None
    after: EventAfterType | None


class HtmxHeaderType(TypedDict, total=False):
    """Type for hx_headers parameter in get_headers()."""

    location: LocationType | None
    redirect: str | None
    refresh: bool
    push_url: PushUrlType | None
    replace_url: PushUrlType | None
    re_swap: ReSwapMethod | None
    re_target: str | None
    trigger_event: TriggerEventType | None



from __future__ import annotations

import pytest

from litestar.dto._backend import DTOBackend
from litestar.dto._codegen_backend import DTOCodegenBackend


@pytest.fixture()
def backend_cls(use_experimental_dto_backend: bool) -> type[DTOBackend | DTOCodegenBackend]:
    return DTOCodegenBackend if use_experimental_dto_backend else DTOBackend



from __future__ import annotations

from typing import TYPE_CHECKING

from msgspec import Meta, Struct, field
from typing_extensions import Annotated

from litestar.dto import MsgspecDTO, dto_field
from litestar.dto.data_structures import DTOFieldDefinition
from litestar.typing import FieldDefinition

if TYPE_CHECKING:
    from typing import Callable


def test_field_definition_generation(
    int_factory: Callable[[], int], expected_field_defs: list[DTOFieldDefinition]
) -> None:
    class TestStruct(Struct):
        a: int
        b: Annotated[int, Meta(extra=dto_field("read-only"))]
        c: Annotated[int, Meta(gt=1)]
        d: int = field(default=1)
        e: int = field(default_factory=int_factory)

    field_defs = list(MsgspecDTO.generate_field_definitions(TestStruct))
    assert field_defs[0].model_name == "TestStruct"
    for field_def, exp in zip(field_defs, expected_field_defs):
        assert field_def == exp


def test_detect_nested_field() -> None:
    class TestStruct(Struct):
        a: int

    class NotStruct:
        pass

    assert MsgspecDTO.detect_nested_field(FieldDefinition.from_annotation(TestStruct)) is True
    assert MsgspecDTO.detect_nested_field(FieldDefinition.from_annotation(NotStruct)) is False



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from litestar.types.empty import Empty

if TYPE_CHECKING:
    from litestar.types.empty import EmptyType

ValueT = TypeVar("ValueT")
DefaultT = TypeVar("DefaultT")


def value_or_default(value: ValueT | EmptyType, default: DefaultT) -> ValueT | DefaultT:
    """Return `value` handling the case where it is empty.

    If `value` is `Empty`, `default` is returned.

    Args:
        value: The value to check.
        default: The default value to return if `value` is `Empty`.

    Returns:
        The value or default value.
    """
    return default if value is Empty else value



from .msgspec_hooks import (
    decode_json,
    decode_msgpack,
    default_deserializer,
    default_serializer,
    encode_json,
    encode_msgpack,
    get_serializer,
)

__all__ = (
    "default_deserializer",
    "decode_json",
    "decode_msgpack",
    "default_serializer",
    "encode_json",
    "encode_msgpack",
    "get_serializer",
)



from __future__ import annotations

from typing import Any

from litestar.utils.deprecation import warn_deprecation

from . import minijinja as _minijinja


def __getattr__(name: str) -> Any:
    warn_deprecation(
        "2.3.0",
        "contrib.minijnja",
        "import",
        removal_in="3.0.0",
        alternative="contrib.minijinja",
    )
    return getattr(_minijinja, name)



from docs.examples.pagination.using_cursor_pagination import app

from litestar.status_codes import HTTP_200_OK
from litestar.testing import TestClient


def test_using_cursor_pagination() -> None:
    with TestClient(app) as client:
        response = client.get("/people", params={"results_per_page": 5})
        assert response.status_code == HTTP_200_OK
        response_data = response.json()
        assert len(response_data["items"]) == 5
        assert response_data["results_per_page"] == 5
        assert isinstance(response_data["cursor"], str)



from .dataclass import DataclassSchemaPlugin
from .pagination import PaginationSchemaPlugin
from .struct import StructSchemaPlugin
from .typed_dict import TypedDictSchemaPlugin

__all__ = ("openapi_schema_plugins",)

# NOTE: The Pagination type plugin has to come before the Dataclass plugin since the Pagination
# classes are dataclasses, but we want to handle them differently from how dataclasses are normally
# handled.
openapi_schema_plugins = [
    PaginationSchemaPlugin(),
    StructSchemaPlugin(),
    DataclassSchemaPlugin(),
    TypedDictSchemaPlugin(),
]



from __future__ import annotations

from asyncio import Queue
from collections import defaultdict, deque
from typing import Any, AsyncGenerator, Iterable

from litestar.channels.backends.base import ChannelsBackend


class MemoryChannelsBackend(ChannelsBackend):
    """An in-memory channels backend"""

    def __init__(self, history: int = 0) -> None:
        self._max_history_length = history
        self._channels: set[str] = set()
        self._queue: Queue[tuple[str, bytes]] | None = None
        self._history: defaultdict[str, deque[bytes]] = defaultdict(lambda: deque(maxlen=self._max_history_length))

    async def on_startup(self) -> None:
        self._queue = Queue()

    async def on_shutdown(self) -> None:
        self._queue = None

    async def publish(self, data: bytes, channels: Iterable[str]) -> None:
        """Publish ``data`` to ``channels``. If a channel has not yet been subscribed to,
        this will be a no-op.

        Args:
            data: Data to publish
            channels: Channels to publish to

        Returns:
            None

        Raises:
            RuntimeError: If ``on_startup`` has not been called yet
        """
        if not self._queue:
            raise RuntimeError("Backend not yet initialized. Did you forget to call on_startup?")

        for channel in channels:
            if channel not in self._channels:
                continue

            self._queue.put_nowait((channel, data))
        if self._max_history_length:
            for channel in channels:
                self._history[channel].append(data)

    async def subscribe(self, channels: Iterable[str]) -> None:
        """Subscribe to ``channels``, and enable publishing to them"""
        self._channels.update(channels)

    async def unsubscribe(self, channels: Iterable[str]) -> None:
        """Unsubscribe from ``channels``"""
        self._channels -= set(channels)
        try:
            for channel in channels:
                del self._history[channel]
        except KeyError:
            pass

    async def stream_events(self) -> AsyncGenerator[tuple[str, Any], None]:
        """Return a generator, iterating over events of subscribed channels as they become available"""
        while self._queue:
            yield await self._queue.get()
            self._queue.task_done()

    async def get_history(self, channel: str, limit: int | None = None) -> list[bytes]:
        """Return the event history of ``channel``, at most ``limit`` entries"""
        history = list(self._history[channel])
        if limit:
            history = history[-limit:]
        return history



from typing import Type

from polyfactory import BaseFactory

from litestar import Controller, Litestar
from litestar._openapi.typescript_converter.converter import (
    convert_openapi_to_typescript,
)


def test_openapi_to_typescript_converter(person_controller: Type[Controller], pet_controller: Type[Controller]) -> None:
    BaseFactory.seed_random(1)
    app = Litestar(route_handlers=[person_controller, pet_controller])
    assert app.openapi_schema

    result = convert_openapi_to_typescript(openapi_schema=app.openapi_schema)
    assert (
        result.write().replace("\t", "	") == "export namespace API {\n"
        "\texport namespace PetOwnerOrPetGetPetsOrOwners {\n"
        "\texport namespace Http200 {\n"
        "\texport type ResponseBody = ({\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies?: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "} | {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional?: null | string;\n"
        "\tpets?: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies?: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "})[];\n"
        "\n"
        "\texport interface ResponseHeaders {\n"
        '\t"x-my-tag"?: string;\n'
        "};\n"
        "};\n"
        "\n"
        "\texport namespace Http406 {\n"
        "\texport type ResponseBody = {\n"
        "\tdetail: string;\n"
        "\textra?: Record<string, unknown> | null | unknown[];\n"
        "\tstatus_code: number;\n"
        "};\n"
        "};\n"
        "};\n"
        "\n"
        "\texport namespace PetPets {\n"
        "\texport namespace Http200 {\n"
        "\texport type ResponseBody = {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies?: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "};\n"
        "};\n"
        "\n"
        "\texport namespace ServiceIdPersonBulkBulkCreatePerson {\n"
        "\texport interface HeaderParameters {\n"
        "\tsecret: string;\n"
        "};\n"
        "\n"
        "\texport namespace Http201 {\n"
        "\texport type ResponseBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional: null | string;\n"
        "\tpets: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "}[];\n"
        "};\n"
        "\n"
        "\texport namespace Http400 {\n"
        "\texport type ResponseBody = {\n"
        "\tdetail: string;\n"
        "\textra?: Record<string, unknown> | null | unknown[];\n"
        "\tstatus_code: number;\n"
        "};\n"
        "};\n"
        "\n"
        "\texport type RequestBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional: null | string;\n"
        "\tpets: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "}[];\n"
        "};\n"
        "\n"
        "\texport namespace ServiceIdPersonBulkBulkPartialUpdatePerson {\n"
        "\texport interface HeaderParameters {\n"
        "\tsecret: string;\n"
        "};\n"
        "\n"
        "\texport namespace Http200 {\n"
        "\texport type ResponseBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional: null | string;\n"
        "\tpets: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "}[];\n"
        "};\n"
        "\n"
        "\texport namespace Http400 {\n"
        "\texport type ResponseBody = {\n"
        "\tdetail: string;\n"
        "\textra?: Record<string, unknown> | null | unknown[];\n"
        "\tstatus_code: number;\n"
        "};\n"
        "};\n"
        "\n"
        "\texport type RequestBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional: null | string;\n"
        "\tpets: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "}[];\n"
        "};\n"
        "\n"
        "\texport namespace ServiceIdPersonBulkBulkUpdatePerson {\n"
        "\texport interface HeaderParameters {\n"
        "\tsecret: string;\n"
        "};\n"
        "\n"
        "\texport namespace Http200 {\n"
        "\texport type ResponseBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional?: null | string;\n"
        "\tpets?: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies?: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "}[];\n"
        "};\n"
        "\n"
        "\texport namespace Http400 {\n"
        "\texport type ResponseBody = {\n"
        "\tdetail: string;\n"
        "\textra?: Record<string, unknown> | null | unknown[];\n"
        "\tstatus_code: number;\n"
        "};\n"
        "};\n"
        "\n"
        "\texport type RequestBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional?: null | string;\n"
        "\tpets?: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies?: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "}[];\n"
        "};\n"
        "\n"
        "\texport namespace ServiceIdPersonCreatePerson {\n"
        "\texport interface HeaderParameters {\n"
        "\tsecret: string;\n"
        "};\n"
        "\n"
        "\texport namespace Http201 {\n"
        "\texport type ResponseBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional?: null | string;\n"
        "\tpets?: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies?: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "};\n"
        "};\n"
        "\n"
        "\texport namespace Http400 {\n"
        "\texport type ResponseBody = {\n"
        "\tdetail: string;\n"
        "\textra?: Record<string, unknown> | null | unknown[];\n"
        "\tstatus_code: number;\n"
        "};\n"
        "};\n"
        "\n"
        "\texport type RequestBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional?: null | string;\n"
        "\tpets?: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies?: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "};\n"
        "};\n"
        "\n"
        "\texport namespace ServiceIdPersonDataclassGetPersonDataclass {\n"
        "\texport namespace Http200 {\n"
        "\texport type ResponseBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional?: null | string;\n"
        "\tpets?: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies?: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "};\n"
        "};\n"
        "};\n"
        "\n"
        "\texport namespace ServiceIdPersonGetPersons {\n"
        "\texport interface CookieParameters {\n"
        "\tvalue: number;\n"
        "};\n"
        "\n"
        "\texport interface HeaderParameters {\n"
        "\tsecret: string;\n"
        "};\n"
        "\n"
        "\texport namespace Http200 {\n"
        "\texport type ResponseBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional?: null | string;\n"
        "\tpets?: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies?: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "}[];\n"
        "};\n"
        "\n"
        "\texport namespace Http400 {\n"
        "\texport type ResponseBody = {\n"
        "\tdetail: string;\n"
        "\textra?: Record<string, unknown> | null | unknown[];\n"
        "\tstatus_code: number;\n"
        "};\n"
        "};\n"
        "\n"
        "\texport interface PathParameters {\n"
        "\tservice_id: number;\n"
        "};\n"
        "\n"
        "\texport interface QueryParameters {\n"
        "\tfrom_date?: null | number | string | string;\n"
        '\tgender?: "A" | "F" | "M" | "O" | ("A" | "F" | "M" | "O")[] | null;\n'
        "\tname?: null | string | string[];\n"
        "\tpage: number;\n"
        "\tpageSize: number;\n"
        "\tto_date?: null | number | string | string;\n"
        "};\n"
        "};\n"
        "\n"
        "\texport namespace ServiceIdPersonPersonIdDeletePerson {\n"
        "\texport namespace Http204 {\n"
        "\texport type ResponseBody = undefined;\n"
        "};\n"
        "\n"
        "\texport namespace Http400 {\n"
        "\texport type ResponseBody = {\n"
        "\tdetail: string;\n"
        "\textra?: Record<string, unknown> | null | unknown[];\n"
        "\tstatus_code: number;\n"
        "};\n"
        "};\n"
        "\n"
        "\texport interface PathParameters {\n"
        "\tperson_id: string;\n"
        "};\n"
        "};\n"
        "\n"
        "\texport namespace ServiceIdPersonPersonIdGetPersonById {\n"
        "\texport namespace Http200 {\n"
        "\texport type ResponseBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional?: null | string;\n"
        "\tpets?: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies?: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "};\n"
        "};\n"
        "\n"
        "\texport namespace Http400 {\n"
        "\texport type ResponseBody = {\n"
        "\tdetail: string;\n"
        "\textra?: Record<string, unknown> | null | unknown[];\n"
        "\tstatus_code: number;\n"
        "};\n"
        "};\n"
        "\n"
        "\texport interface PathParameters {\n"
        "\tperson_id: string;\n"
        "};\n"
        "};\n"
        "\n"
        "\texport namespace ServiceIdPersonPersonIdPartialUpdatePerson {\n"
        "\texport namespace Http200 {\n"
        "\texport type ResponseBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional: null | string;\n"
        "\tpets: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "};\n"
        "};\n"
        "\n"
        "\texport namespace Http400 {\n"
        "\texport type ResponseBody = {\n"
        "\tdetail: string;\n"
        "\textra?: Record<string, unknown> | null | unknown[];\n"
        "\tstatus_code: number;\n"
        "};\n"
        "};\n"
        "\n"
        "\texport interface PathParameters {\n"
        "\tperson_id: string;\n"
        "};\n"
        "\n"
        "\texport type RequestBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional: null | string;\n"
        "\tpets: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "};\n"
        "};\n"
        "\n"
        "\texport namespace ServiceIdPersonPersonIdUpdatePerson {\n"
        "\texport namespace Http200 {\n"
        "\texport type ResponseBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional?: null | string;\n"
        "\tpets?: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies?: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "};\n"
        "};\n"
        "\n"
        "\texport namespace Http400 {\n"
        "\texport type ResponseBody = {\n"
        "\tdetail: string;\n"
        "\textra?: Record<string, unknown> | null | unknown[];\n"
        "\tstatus_code: number;\n"
        "};\n"
        "};\n"
        "\n"
        "\texport interface PathParameters {\n"
        "\tperson_id: string;\n"
        "};\n"
        "\n"
        "\texport type RequestBody = {\n"
        "\tcomplex: {\n"
        "\t\n"
        "};\n"
        "\tfirst_name: string;\n"
        "\tid: string;\n"
        "\tlast_name: string;\n"
        "\toptional?: null | string;\n"
        "\tpets?: null | {\n"
        "\tage: number;\n"
        "\tname: string;\n"
        '\tspecies?: "Cat" | "Dog" | "Monkey" | "Pig";\n'
        "}[];\n"
        "};\n"
        "};\n"
        "};"
    )



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    Literal,
    Mapping,
    MutableMapping,
    Sequence,
    Tuple,
    Type,
    Union,
)

__all__ = (
    "Dependencies",
    "ExceptionHandlersMap",
    "Middleware",
    "ParametersMap",
    "PathType",
    "ResponseCookies",
    "ResponseHeaders",
    "Scopes",
    "TypeEncodersMap",
)


if TYPE_CHECKING:
    from os import PathLike
    from pathlib import Path

    from typing_extensions import TypeAlias

    from litestar.datastructures.cookie import Cookie
    from litestar.datastructures.response_header import ResponseHeader
    from litestar.di import Provide
    from litestar.enums import ScopeType
    from litestar.middleware.base import DefineMiddleware, MiddlewareProtocol
    from litestar.params import ParameterKwarg

    from .asgi_types import ASGIApp
    from .callable_types import AnyCallable, ExceptionHandler

Dependencies: TypeAlias = "Mapping[str, Union[Provide, AnyCallable]]"
ExceptionHandlersMap: TypeAlias = "MutableMapping[Union[int, Type[Exception]], ExceptionHandler]"
Middleware: TypeAlias = "Union[Callable[..., ASGIApp], DefineMiddleware, Iterator[Tuple[ASGIApp, Dict[str, Any]]], Type[MiddlewareProtocol]]"
ParametersMap: TypeAlias = "Mapping[str, ParameterKwarg]"
PathType: TypeAlias = "Union[Path, PathLike, str]"
ResponseCookies: TypeAlias = "Union[Sequence[Cookie], Mapping[str, str]]"
ResponseHeaders: TypeAlias = "Union[Sequence[ResponseHeader], Mapping[str, str]]"
Scopes: TypeAlias = "set[Literal[ScopeType.HTTP, ScopeType.WEBSOCKET]]"
TypeDecodersSequence: TypeAlias = "Sequence[tuple[Callable[[Any], bool], Callable[[Any, Any], Any]]]"
TypeEncodersMap: TypeAlias = "Mapping[Any, Callable[[Any], Any]]"



from __future__ import annotations

from advanced_alchemy.types import GUID, ORA_JSONB, BigIntIdentity, DateTimeUTC, JsonB

__all__ = (
    "GUID",
    "ORA_JSONB",
    "DateTimeUTC",
    "BigIntIdentity",
    "JsonB",
)



from litestar import delete
from litestar.testing import create_test_client


def test_handler_return_none_and_204_status_response_empty() -> None:
    @delete(path="/")
    async def route() -> None:
        return None

    with create_test_client(route_handlers=[route]) as client:
        response = client.delete("/")
        assert not response.content



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from litestar.openapi.spec.base import BaseSchemaObject

if TYPE_CHECKING:
    from litestar.openapi.spec.encoding import Encoding
    from litestar.openapi.spec.example import Example
    from litestar.openapi.spec.reference import Reference
    from litestar.openapi.spec.schema import Schema

__all__ = ("OpenAPIMediaType",)


@dataclass
class OpenAPIMediaType(BaseSchemaObject):
    """Each Media Type Object provides schema and examples for the media type identified by its key."""

    schema: Reference | Schema | None = None
    """The schema defining the content of the request, response, or parameter."""

    example: Any | None = None
    """Example of the media type.

    The example object SHOULD be in the correct format as specified by the media type.

    The ``example`` field is mutually exclusive of the ``examples`` field.

    Furthermore, if referencing a ``schema`` which contains an example, the ``example`` value SHALL _override_ the
    example provided by the schema.
    """

    examples: dict[str, Example | Reference] | None = None
    """Examples of the media type.

    Each example object SHOULD match the media type and specified schema if present.

    The ``examples`` field is mutually exclusive of the ``example`` field.

    Furthermore, if referencing a ``schema`` which contains an example, the ``examples`` value SHALL _override_ the
    example provided by the schema.
    """

    encoding: dict[str, Encoding] | None = None
    """A map between a property name and its encoding information.

    The key, being the property name, MUST exist in the schema as a property. The encoding object SHALL only apply to
    ``requestBody`` objects when the media type is ``multipart`` or ``application/x-www-form-urlencoded``.
    """



from pathlib import Path
from typing import Dict

import pytest

from litestar import HttpMethod, Litestar, WebSocket, delete, get, route
from litestar.exceptions import ImproperlyConfiguredException, ValidationException
from litestar.handlers.http_handlers import HTTPRouteHandler
from litestar.response import File, Redirect
from litestar.status_codes import (
    HTTP_100_CONTINUE,
    HTTP_200_OK,
    HTTP_304_NOT_MODIFIED,
    HTTP_307_TEMPORARY_REDIRECT,
)
from tests.models import DataclassPerson


def test_route_handler_validation_http_method() -> None:
    # doesn't raise for http methods
    for value in (*list(HttpMethod), *[x.upper() for x in list(HttpMethod)]):
        assert route(http_method=value)  # type: ignore

    # raises for invalid values
    with pytest.raises(ValidationException):
        HTTPRouteHandler(http_method="deleze")  # type: ignore

    # also when passing an empty list
    with pytest.raises(ImproperlyConfiguredException):
        route(http_method=[], status_code=HTTP_200_OK)

    # also when passing malformed tokens
    with pytest.raises(ValidationException):
        route(http_method=[HttpMethod.GET, "poft"], status_code=HTTP_200_OK)  # type: ignore


async def test_function_validation() -> None:
    with pytest.raises(ImproperlyConfiguredException):

        @get(path="/")
        def method_with_no_annotation():  # type: ignore
            pass

        Litestar(route_handlers=[method_with_no_annotation])

        method_with_no_annotation.on_registration(Litestar())

    with pytest.raises(ImproperlyConfiguredException):

        @delete(path="/")
        def method_with_no_content() -> Dict[str, str]:
            return {}

        Litestar(route_handlers=[method_with_no_content])

        method_with_no_content.on_registration(Litestar())

    with pytest.raises(ImproperlyConfiguredException):

        @get(path="/", status_code=HTTP_304_NOT_MODIFIED)
        def method_with_not_modified() -> Dict[str, str]:
            return {}

        Litestar(route_handlers=[method_with_not_modified])

        method_with_not_modified.on_registration(Litestar())

    with pytest.raises(ImproperlyConfiguredException):

        @get(path="/", status_code=HTTP_100_CONTINUE)
        def method_with_status_lower_than_200() -> Dict[str, str]:
            return {}

        Litestar(route_handlers=[method_with_status_lower_than_200])

        method_with_status_lower_than_200.on_registration(Litestar())

    @get(path="/", status_code=HTTP_307_TEMPORARY_REDIRECT)
    def redirect_method() -> Redirect:
        return Redirect("/test")

    Litestar(route_handlers=[redirect_method])

    redirect_method.on_registration(Litestar())

    @get(path="/")
    def file_method() -> File:
        return File(path=Path("."), filename="test_validations.py")

    Litestar(route_handlers=[file_method])

    file_method.on_registration(Litestar())

    assert not file_method.media_type

    with pytest.raises(ImproperlyConfiguredException):

        @get(path="/test")
        def test_function_1(socket: WebSocket) -> None:
            return None

        test_function_1.on_registration(Litestar())

    with pytest.raises(ImproperlyConfiguredException):

        @get("/person")
        def test_function_2(self, data: DataclassPerson) -> None:  # type: ignore
            return None

        Litestar(route_handlers=[test_function_2])

        test_function_2.on_registration(Litestar())



from __future__ import annotations

import secrets
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal

from litestar.datastructures import Cookie, MutableScopeHeaders
from litestar.enums import ScopeType
from litestar.exceptions import ImproperlyConfiguredException
from litestar.middleware.session.base import ONE_DAY_IN_SECONDS, BaseBackendConfig, BaseSessionBackend
from litestar.types import Empty, Message, Scopes, ScopeSession
from litestar.utils.dataclass import extract_dataclass_items

__all__ = ("ServerSideSessionBackend", "ServerSideSessionConfig")


if TYPE_CHECKING:
    from litestar import Litestar
    from litestar.connection import ASGIConnection
    from litestar.stores.base import Store


class ServerSideSessionBackend(BaseSessionBackend["ServerSideSessionConfig"]):
    """Base class for server-side backends.

    Implements :class:`BaseSessionBackend` and defines and interface which subclasses can
    implement to facilitate the storage of session data.
    """

    def __init__(self, config: ServerSideSessionConfig) -> None:
        """Initialize ``ServerSideSessionBackend``

        Args:
            config: A subclass of ``ServerSideSessionConfig``
        """
        super().__init__(config=config)

    async def get(self, session_id: str, store: Store) -> bytes | None:
        """Retrieve data associated with ``session_id``.

        Args:
            session_id: The session-ID
            store: Store to retrieve the session data from

        Returns:
            The session data, if existing, otherwise ``None``.
        """
        max_age = int(self.config.max_age) if self.config.max_age is not None else None
        return await store.get(session_id, renew_for=max_age if self.config.renew_on_access else None)

    async def set(self, session_id: str, data: bytes, store: Store) -> None:
        """Store ``data`` under the ``session_id`` for later retrieval.

        If there is already data associated with ``session_id``, replace
        it with ``data`` and reset its expiry time

        Args:
            session_id: The session-ID
            data: Serialized session data
            store: Store to save the session data in

        Returns:
            None
        """
        expires_in = int(self.config.max_age) if self.config.max_age is not None else None
        await store.set(session_id, data, expires_in=expires_in)

    async def delete(self, session_id: str, store: Store) -> None:
        """Delete the data associated with ``session_id``. Fails silently if no such session-ID exists.

        Args:
            session_id: The session-ID
            store: Store to delete the session data from

        Returns:
            None
        """
        await store.delete(session_id)

    def generate_session_id(self) -> str:
        """Generate a new session-ID, with
        n=:attr:`session_id_bytes <ServerSideSessionConfig.session_id_bytes>` random bytes.

        Returns:
            A session-ID
        """
        return secrets.token_hex(self.config.session_id_bytes)

    async def store_in_message(self, scope_session: ScopeSession, message: Message, connection: ASGIConnection) -> None:
        """Store the necessary information in the outgoing ``Message`` by setting a cookie containing the session-ID.

        If the session is empty, a null-cookie will be set. Otherwise, the serialised
        data will be stored using :meth:`set <ServerSideSessionBackend.set>`, under the current session-id. If no session-ID
        exists, a new ID will be generated using :meth:`generate_session_id <ServerSideSessionBackend.generate_session_id>`.

        Args:
            scope_session: Current session to store
            message: Outgoing send-message
            connection: Originating ASGIConnection containing the scope

        Returns:
            None
        """
        scope = connection.scope
        store = self.config.get_store_from_app(scope["app"])
        headers = MutableScopeHeaders.from_message(message)
        session_id = connection.cookies.get(self.config.key)
        if not session_id or session_id == "null":
            session_id = self.generate_session_id()

        cookie_params = dict(extract_dataclass_items(self.config, exclude_none=True, include=Cookie.__dict__.keys()))

        if scope_session is Empty:
            await self.delete(session_id, store=store)
            headers.add(
                "Set-Cookie",
                Cookie(value="null", key=self.config.key, expires=0, **cookie_params).to_header(header=""),
            )
        else:
            serialised_data = self.serialize_data(scope_session, scope)
            await self.set(session_id=session_id, data=serialised_data, store=store)
            headers.add(
                "Set-Cookie", Cookie(value=session_id, key=self.config.key, **cookie_params).to_header(header="")
            )

    async def load_from_connection(self, connection: ASGIConnection) -> dict[str, Any]:
        """Load session data from a connection and return it as a dictionary to be used in the current application
        scope.

        The session-ID will be gathered from a cookie with the key set in
        :attr:`BaseBackendConfig.key`. If a cookie is found, its value will be used as the session-ID and data associated
        with this ID will be loaded using :meth:`get <ServerSideSessionBackend.get>`.
        If no cookie was found or no data was loaded from the store, this will return an
        empty dictionary.

        Args:
            connection: An ASGIConnection instance

        Returns:
            The current session data
        """
        if session_id := connection.cookies.get(self.config.key):
            store = self.config.get_store_from_app(connection.scope["app"])
            data = await self.get(session_id, store=store)
            if data is not None:
                return self.deserialize_data(data)
        return {}


@dataclass
class ServerSideSessionConfig(BaseBackendConfig[ServerSideSessionBackend]):
    """Base configuration for server side backends."""

    _backend_class = ServerSideSessionBackend

    session_id_bytes: int = field(default=32)
    """Number of bytes used to generate a random session-ID."""
    renew_on_access: bool = field(default=False)
    """Renew expiry times of sessions when they're being accessed"""
    key: str = field(default="session")
    """Key to use for the cookie inside the header, e.g. ``session=<data>`` where ``session`` is the cookie key and
    ``<data>`` is the session data.

    Notes:
        - If a session cookie exceeds 4KB in size it is split. In this case the key will be of the format
          ``session-{segment number}``.

    """
    max_age: int = field(default=ONE_DAY_IN_SECONDS * 14)
    """Maximal age of the cookie before its invalidated."""
    scopes: Scopes = field(default_factory=lambda: {ScopeType.HTTP, ScopeType.WEBSOCKET})
    """Scopes for the middleware - options are ``http`` and ``websocket`` with the default being both"""
    path: str = field(default="/")
    """Path fragment that must exist in the request url for the cookie to be valid.

    Defaults to ``'/'``.
    """
    domain: str | None = field(default=None)
    """Domain for which the cookie is valid."""
    secure: bool = field(default=False)
    """Https is required for the cookie."""
    httponly: bool = field(default=True)
    """Forbids javascript to access the cookie via 'Document.cookie'."""
    samesite: Literal["lax", "strict", "none"] = field(default="lax")
    """Controls whether or not a cookie is sent with cross-site requests. Defaults to ``lax``."""
    exclude: str | list[str] | None = field(default=None)
    """A pattern or list of patterns to skip in the session middleware."""
    exclude_opt_key: str = field(default="skip_session")
    """An identifier to use on routes to disable the session middleware for a particular route."""
    store: str = "sessions"
    """Name of the :class:`Store <.stores.base.Store>` to use"""

    def __post_init__(self) -> None:
        if len(self.key) < 1 or len(self.key) > 256:
            raise ImproperlyConfiguredException("key must be a string with a length between 1-256")
        if self.max_age < 1:
            raise ImproperlyConfiguredException("max_age must be greater than 0")

    def get_store_from_app(self, app: Litestar) -> Store:
        """Get the store defined in :attr:`store` from an :class:`Litestar <.app.Litestar>` instance"""
        return app.stores.get(self.store)



# ruff: noqa: UP006
from __future__ import annotations

import inspect
import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from dataclasses import fields
from typing import TYPE_CHECKING, List, Tuple
from unittest.mock import MagicMock, Mock, PropertyMock

import pytest
from click import Group
from pytest import MonkeyPatch

from litestar import Litestar, MediaType, Request, Response, get
from litestar.config.app import AppConfig
from litestar.config.response_cache import ResponseCacheConfig
from litestar.contrib.sqlalchemy.plugins import SQLAlchemySerializationPlugin
from litestar.datastructures import MutableScopeHeaders, State
from litestar.events.emitter import SimpleEventEmitter
from litestar.exceptions import (
    ImproperlyConfiguredException,
    InternalServerException,
    LitestarWarning,
    NotFoundException,
)
from litestar.logging.config import LoggingConfig
from litestar.plugins import CLIPluginProtocol
from litestar.router import Router
from litestar.status_codes import HTTP_200_OK, HTTP_400_BAD_REQUEST, HTTP_500_INTERNAL_SERVER_ERROR
from litestar.testing import TestClient, create_test_client

if TYPE_CHECKING:
    from typing import Dict

    from litestar.types import Message, Scope


@pytest.fixture()
def app_config_object() -> AppConfig:
    return AppConfig(
        after_exception=[],
        after_request=None,
        after_response=None,
        allowed_hosts=[],
        before_request=None,
        before_send=[],
        response_cache_config=ResponseCacheConfig(),
        cache_control=None,
        compression_config=None,
        cors_config=None,
        csrf_config=None,
        debug=False,
        dependencies={},
        etag=None,
        event_emitter_backend=SimpleEventEmitter,
        exception_handlers={},
        guards=[],
        listeners=[],
        logging_config=None,
        middleware=[],
        multipart_form_part_limit=1000,
        on_shutdown=[],
        on_startup=[],
        openapi_config=None,
        opt={},
        parameters={},
        plugins=[],
        request_class=None,
        response_class=None,
        response_cookies=[],
        response_headers=[],
        route_handlers=[],
        security=[],
        static_files_config=[],
        tags=[],
        template_config=None,
        websocket_class=None,
    )


def test_access_openapi_schema_raises_if_not_configured() -> None:
    """Test that accessing the openapi schema raises if not configured."""
    app = Litestar(openapi_config=None)
    with pytest.raises(ImproperlyConfiguredException):
        app.openapi_schema


def test_set_debug_updates_logging_level() -> None:
    app = Litestar()

    assert app.logger is not None
    assert app.logger.level == logging.INFO  # type: ignore[attr-defined]

    app.debug = True
    assert app.logger.level == logging.DEBUG  # type: ignore[attr-defined]

    app.debug = False
    assert app.logger.level == logging.INFO  # type: ignore[attr-defined]


@pytest.mark.parametrize("env_name,app_attr", [("LITESTAR_DEBUG", "debug"), ("LITESTAR_PDB", "pdb_on_exception")])
@pytest.mark.parametrize(
    "env_value,app_value,expected_value",
    [
        (None, None, False),
        (None, False, False),
        (None, True, True),
        ("0", None, False),
        ("0", False, False),
        ("0", True, True),
        ("1", None, True),
        ("1", False, False),
        ("1", True, True),
    ],
)
@pytest.mark.filterwarnings("ignore::litestar.utils.warnings.LitestarWarning:")
def test_set_env_flags(
    monkeypatch: MonkeyPatch,
    env_value: str | None,
    app_value: bool | None,
    expected_value: bool,
    env_name: str,
    app_attr: str,
) -> None:
    if env_value is not None:
        monkeypatch.setenv(env_name, env_value)
    else:
        monkeypatch.delenv(env_name, raising=False)

    app = Litestar(**{app_attr: app_value})  # type: ignore[arg-type]

    assert getattr(app, app_attr) is expected_value


def test_warn_pdb_on_exception() -> None:
    with pytest.warns(LitestarWarning, match="Debugger"):
        Litestar(pdb_on_exception=True)


def test_app_params_defined_on_app_config_object() -> None:
    """Ensures that all parameters to the `Litestar` constructor are present on the `AppConfig` object."""
    litestar_signature = inspect.signature(Litestar)
    app_config_fields = {f.name for f in fields(AppConfig)}
    for name in litestar_signature.parameters:
        if name in {"on_app_init", "initial_state", "_preferred_validation_backend"}:
            continue
        assert name in app_config_fields
    # ensure there are not fields defined on AppConfig that aren't in the Litestar signature
    assert not (app_config_fields - set(litestar_signature.parameters.keys()))


def test_app_config_object_used(app_config_object: AppConfig, monkeypatch: pytest.MonkeyPatch) -> None:
    """Ensure that the properties on the `AppConfig` object are accessed within the `Litestar` constructor.

    In the test we replace every field on the `AppConfig` type with a property mock so that we can check that it has at
    least been accessed. It doesn't actually check that we do the right thing with it, but is a guard against the case
    of adding a parameter to the `Litestar` signature and to the `AppConfig` object, and using the value from the
    parameter downstream from construction of the `AppConfig` object.
    """

    # replace each field on the `AppConfig` object with a `PropertyMock`, this allows us to assert that the properties
    # have been accessed during app instantiation.
    property_mocks: List[Tuple[str, Mock]] = []
    for field in fields(AppConfig):
        if field.name == "response_cache_config":
            property_mock = PropertyMock(return_value=ResponseCacheConfig())
        if field.name in ["event_emitter_backend", "response_cache_config"]:
            property_mock = PropertyMock(return_value=Mock())
        else:
            # default iterable return value allows the mock properties that need to be iterated over in
            # `Litestar.__init__()` to not blow up, for other properties it shouldn't matter what the value is for the
            # sake of this test.
            property_mock = PropertyMock(return_value=[])
        property_mocks.append((field.name, property_mock))
        monkeypatch.setattr(type(app_config_object), field.name, property_mock, raising=False)

    # Things that we don't actually need to call for this test
    monkeypatch.setattr(Litestar, "register", MagicMock())
    monkeypatch.setattr(Litestar, "_create_asgi_handler", MagicMock())
    monkeypatch.setattr(Router, "__init__", MagicMock())

    # instantiates the app with an `on_app_config` that returns our patched `AppConfig` object.
    Litestar(on_app_init=[MagicMock(return_value=app_config_object)])

    # this ensures that each of the properties of the `AppConfig` object have been accessed within `Litestar.__init__()`
    for name, mock in property_mocks:
        assert mock.called, f"expected {name} to be called"


def test_app_debug_create_logger() -> None:
    app = Litestar([], debug=True)

    assert app.logging_config
    assert app.logging_config.loggers["litestar"]["level"] == "DEBUG"  # type: ignore[attr-defined]


def test_app_debug_explicitly_disable_logging() -> None:
    app = Litestar([], logging_config=None)

    assert not app.logging_config


def test_app_debug_update_logging_config() -> None:
    logging_config = LoggingConfig()
    app = Litestar([], logging_config=logging_config, debug=True)

    assert app.logging_config is logging_config
    assert app.logging_config.loggers["litestar"]["level"] == "DEBUG"  # type: ignore[attr-defined]


def test_set_state() -> None:
    def modify_state_in_hook(app_config: AppConfig) -> AppConfig:
        assert isinstance(app_config.state, State)
        app_config.state["c"] = "D"
        app_config.state["e"] = "f"
        return app_config

    app = Litestar(state=State({"a": "b", "c": "d"}), on_app_init=[modify_state_in_hook])
    assert app.state._state == {"a": "b", "c": "D", "e": "f"}


def test_app_from_config(app_config_object: AppConfig) -> None:
    Litestar.from_config(app_config_object)


def test_before_send() -> None:
    @get("/test")
    def handler() -> Dict[str, str]:
        return {"key": "value"}

    async def before_send_hook_handler(message: Message, scope: Scope) -> None:
        if message["type"] == "http.response.start":
            headers = MutableScopeHeaders(message)
            headers.add("My Header", scope["app"].state.message)

    def on_startup(app: Litestar) -> None:
        app.state.message = "value injected during send"

    with create_test_client(handler, on_startup=[on_startup], before_send=[before_send_hook_handler]) as client:
        response = client.get("/test")
        assert response.status_code == HTTP_200_OK
        assert response.headers.get("My Header") == "value injected during send"


def test_using_custom_http_exception_handler() -> None:
    @get("/{param:int}")
    def my_route_handler(param: int) -> None:
        ...

    def my_custom_handler(_: Request, __: Exception) -> Response:
        return Response(content="custom message", media_type=MediaType.TEXT, status_code=HTTP_400_BAD_REQUEST)

    with create_test_client(my_route_handler, exception_handlers={NotFoundException: my_custom_handler}) as client:
        response = client.get("/abc")
        assert response.text == "custom message"
        assert response.status_code == HTTP_400_BAD_REQUEST


def test_debug_response_created() -> None:
    # this will test exception causes are recorded in output
    # since frames include code in context we should not raise
    # exception directly
    def exception_thrower() -> float:
        return 1 / 0

    @get("/")
    def my_route_handler() -> None:
        try:
            exception_thrower()
        except Exception as e:
            raise InternalServerException() from e

    app = Litestar(route_handlers=[my_route_handler], debug=True)
    client = TestClient(app=app)

    response = client.get("/")
    assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
    assert "text/plain" in response.headers["content-type"]

    response = client.get("/", headers={"accept": "text/html"})
    assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR
    assert "text/html" in response.headers["content-type"]
    assert "ZeroDivisionError" in response.text


def test_handler_error_return_status_500() -> None:
    @get("/")
    def my_route_handler() -> None:
        raise KeyError("custom message")

    with create_test_client(my_route_handler) as client:
        response = client.get("/")
        assert response.status_code == HTTP_500_INTERNAL_SERVER_ERROR


def test_lifespan() -> None:
    events: list[str] = []
    counter = {"value": 0}

    def sync_function_without_app() -> None:
        events.append("sync_function_without_app")
        counter["value"] += 1

    async def async_function_without_app() -> None:
        events.append("async_function_without_app")
        counter["value"] += 1

    def sync_function_with_app(app: Litestar) -> None:
        events.append("sync_function_with_app")
        assert app is not None
        assert isinstance(app.state, State)
        counter["value"] += 1
        app.state.x = True

    async def async_function_with_app(app: Litestar) -> None:
        events.append("async_function_with_app")
        assert app is not None
        assert isinstance(app.state, State)
        counter["value"] += 1
        app.state.y = True

    with create_test_client(
        [],
        on_startup=[
            sync_function_without_app,
            async_function_without_app,
            sync_function_with_app,
            async_function_with_app,
        ],
        on_shutdown=[
            sync_function_without_app,
            async_function_without_app,
            sync_function_with_app,
            async_function_with_app,
        ],
    ) as client:
        assert counter["value"] == 4
        assert client.app.state.x
        assert client.app.state.y
        counter["value"] = 0
        assert counter["value"] == 0
    assert counter["value"] == 4
    assert events == [
        "sync_function_without_app",
        "async_function_without_app",
        "sync_function_with_app",
        "async_function_with_app",
        "sync_function_without_app",
        "async_function_without_app",
        "sync_function_with_app",
        "async_function_with_app",
    ]


def test_registering_route_handler_generates_openapi_docs() -> None:
    def fn() -> None:
        return

    app = Litestar(route_handlers=[])
    assert app.openapi_schema

    app.register(get("/path1")(fn))

    assert app.openapi_schema.paths is not None
    assert app.openapi_schema.paths.get("/path1")

    app.register(get("/path2")(fn))
    assert app.openapi_schema.paths.get("/path1")
    assert app.openapi_schema.paths.get("/path2")


def test_plugin_properties() -> None:
    class FooPlugin(CLIPluginProtocol):
        def on_cli_init(self, cli: Group) -> None:
            return

    app = Litestar(plugins=[FooPlugin(), SQLAlchemySerializationPlugin()])

    assert app.openapi_schema_plugins == list(app.plugins.openapi)
    assert app.cli_plugins == list(app.plugins.cli)
    assert app.serialization_plugins == list(app.plugins.serialization)


def test_plugin_registry() -> None:
    class FooPlugin(CLIPluginProtocol):
        def on_cli_init(self, cli: Group) -> None:
            return

    foo = FooPlugin()

    app = Litestar(plugins=[foo])

    assert foo in app.plugins.cli


def test_lifespan_context_and_shutdown_hook_execution_order() -> None:
    events: list[str] = []
    counter = {"value": 0}

    @asynccontextmanager
    async def lifespan_context_1(app: Litestar) -> AsyncGenerator[None, None]:
        try:
            yield
        finally:
            events.append("ctx_1")
            counter["value"] += 1

    @asynccontextmanager
    async def lifespan_context_2(app: Litestar) -> AsyncGenerator[None, None]:
        try:
            yield
        finally:
            events.append("ctx_2")
            counter["value"] += 1

    async def hook_a(app: Litestar) -> None:
        events.append("hook_a")
        counter["value"] += 1

    async def hook_b(app: Litestar) -> None:
        events.append("hook_b")
        counter["value"] += 1

    with create_test_client(
        route_handlers=[],
        lifespan=[
            lifespan_context_1,
            lifespan_context_2,
        ],
        on_shutdown=[hook_a, hook_b],
    ):
        assert counter["value"] == 0

    assert counter["value"] == 4
    assert events[0] == "ctx_2"
    assert events[1] == "ctx_1"
    assert events[2] == "hook_a"
    assert events[3] == "hook_b"



from __future__ import annotations

from typing import TYPE_CHECKING

from litestar.datastructures import Headers, MutableScopeHeaders
from litestar.enums import ScopeType
from litestar.middleware.base import AbstractMiddleware

__all__ = ("CORSMiddleware",)


if TYPE_CHECKING:
    from litestar.config.cors import CORSConfig
    from litestar.types import ASGIApp, Message, Receive, Scope, Send


class CORSMiddleware(AbstractMiddleware):
    """CORS Middleware."""

    __slots__ = ("config",)

    def __init__(self, app: ASGIApp, config: CORSConfig) -> None:
        """Middleware that adds CORS validation to the application.

        Args:
            app: The ``next`` ASGI app to call.
            config: An instance of :class:`CORSConfig <litestar.config.cors.CORSConfig>`
        """
        super().__init__(app=app, scopes={ScopeType.HTTP})
        self.config = config

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """ASGI callable.

        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """
        headers = Headers.from_scope(scope=scope)
        if origin := headers.get("origin"):
            await self.app(scope, receive, self.send_wrapper(send=send, origin=origin, has_cookie="cookie" in headers))
        else:
            await self.app(scope, receive, send)

    def send_wrapper(self, send: Send, origin: str, has_cookie: bool) -> Send:
        """Wrap ``send`` to ensure that state is not disconnected.

        Args:
            has_cookie: Boolean flag dictating if the connection has a cookie set.
            origin: The value of the ``Origin`` header.
            send: The ASGI send function.

        Returns:
            An ASGI send function.
        """

        async def wrapped_send(message: Message) -> None:
            if message["type"] == "http.response.start":
                message.setdefault("headers", [])
                headers = MutableScopeHeaders.from_message(message=message)
                headers.update(self.config.simple_headers)

                if (self.config.is_allow_all_origins and has_cookie) or (
                    not self.config.is_allow_all_origins and self.config.is_origin_allowed(origin=origin)
                ):
                    headers["Access-Control-Allow-Origin"] = origin
                    headers["Vary"] = "Origin"

            await send(message)

        return wrapped_send



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Literal, Mapping, Sequence

from litestar.app import DEFAULT_OPENAPI_CONFIG, Litestar
from litestar.controller import Controller
from litestar.events import SimpleEventEmitter
from litestar.testing.client import AsyncTestClient, TestClient
from litestar.types import Empty
from litestar.utils.predicates import is_class_and_subclass

if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager

    from litestar import Request, WebSocket
    from litestar.config.allowed_hosts import AllowedHostsConfig
    from litestar.config.app import ExperimentalFeatures
    from litestar.config.compression import CompressionConfig
    from litestar.config.cors import CORSConfig
    from litestar.config.csrf import CSRFConfig
    from litestar.config.response_cache import ResponseCacheConfig
    from litestar.datastructures import CacheControlHeader, ETag, ResponseHeader, State
    from litestar.dto import AbstractDTO
    from litestar.events import BaseEventEmitterBackend, EventListener
    from litestar.logging.config import BaseLoggingConfig
    from litestar.middleware.session.base import BaseBackendConfig
    from litestar.openapi.config import OpenAPIConfig
    from litestar.openapi.spec import SecurityRequirement
    from litestar.plugins import PluginProtocol
    from litestar.static_files.config import StaticFilesConfig
    from litestar.stores.base import Store
    from litestar.stores.registry import StoreRegistry
    from litestar.template.config import TemplateConfig
    from litestar.types import (
        AfterExceptionHookHandler,
        AfterRequestHookHandler,
        AfterResponseHookHandler,
        BeforeMessageSendHookHandler,
        BeforeRequestHookHandler,
        ControllerRouterHandler,
        Dependencies,
        EmptyType,
        ExceptionHandlersMap,
        Guard,
        LifespanHook,
        Middleware,
        OnAppInitHandler,
        ParametersMap,
        ResponseCookies,
        ResponseType,
        TypeEncodersMap,
    )


def create_test_client(
    route_handlers: ControllerRouterHandler | Sequence[ControllerRouterHandler] | None = None,
    *,
    after_exception: Sequence[AfterExceptionHookHandler] | None = None,
    after_request: AfterRequestHookHandler | None = None,
    after_response: AfterResponseHookHandler | None = None,
    allowed_hosts: Sequence[str] | AllowedHostsConfig | None = None,
    backend: Literal["asyncio", "trio"] = "asyncio",
    backend_options: Mapping[str, Any] | None = None,
    base_url: str = "http://testserver.local",
    before_request: BeforeRequestHookHandler | None = None,
    before_send: Sequence[BeforeMessageSendHookHandler] | None = None,
    cache_control: CacheControlHeader | None = None,
    compression_config: CompressionConfig | None = None,
    cors_config: CORSConfig | None = None,
    csrf_config: CSRFConfig | None = None,
    debug: bool = True,
    dependencies: Dependencies | None = None,
    dto: type[AbstractDTO] | None | EmptyType = Empty,
    etag: ETag | None = None,
    event_emitter_backend: type[BaseEventEmitterBackend] = SimpleEventEmitter,
    exception_handlers: ExceptionHandlersMap | None = None,
    guards: Sequence[Guard] | None = None,
    include_in_schema: bool | EmptyType = Empty,
    listeners: Sequence[EventListener] | None = None,
    logging_config: BaseLoggingConfig | EmptyType | None = Empty,
    middleware: Sequence[Middleware] | None = None,
    multipart_form_part_limit: int = 1000,
    on_app_init: Sequence[OnAppInitHandler] | None = None,
    on_shutdown: Sequence[LifespanHook] | None = None,
    on_startup: Sequence[LifespanHook] | None = None,
    openapi_config: OpenAPIConfig | None = DEFAULT_OPENAPI_CONFIG,
    opt: Mapping[str, Any] | None = None,
    parameters: ParametersMap | None = None,
    plugins: Sequence[PluginProtocol] | None = None,
    lifespan: list[Callable[[Litestar], AbstractAsyncContextManager] | AbstractAsyncContextManager] | None = None,
    raise_server_exceptions: bool = True,
    pdb_on_exception: bool | None = None,
    request_class: type[Request] | None = None,
    response_cache_config: ResponseCacheConfig | None = None,
    response_class: ResponseType | None = None,
    response_cookies: ResponseCookies | None = None,
    response_headers: Sequence[ResponseHeader] | None = None,
    return_dto: type[AbstractDTO] | None | EmptyType = Empty,
    root_path: str = "",
    security: Sequence[SecurityRequirement] | None = None,
    session_config: BaseBackendConfig | None = None,
    signature_namespace: Mapping[str, Any] | None = None,
    signature_types: Sequence[Any] | None = None,
    state: State | None = None,
    static_files_config: Sequence[StaticFilesConfig] | None = None,
    stores: StoreRegistry | dict[str, Store] | None = None,
    tags: Sequence[str] | None = None,
    template_config: TemplateConfig | None = None,
    timeout: float | None = None,
    type_encoders: TypeEncodersMap | None = None,
    websocket_class: type[WebSocket] | None = None,
    experimental_features: list[ExperimentalFeatures] | None = None,
) -> TestClient[Litestar]:
    """Create a Litestar app instance and initializes it.

    :class:`TestClient <litestar.testing.TestClient>` with it.

    Notes:
        - This function should be called as a context manager to ensure async startup and shutdown are
            handled correctly.

    Examples:
        .. code-block:: python

            from litestar import get
            from litestar.testing import create_test_client


            @get("/some-path")
            def my_handler() -> dict[str, str]:
                return {"hello": "world"}


            def test_my_handler() -> None:
                with create_test_client(my_handler) as client:
                    response == client.get("/some-path")
                    assert response.json() == {"hello": "world"}

    Args:
        route_handlers: A single handler or a sequence of route handlers, which can include instances of
            :class:`Router <litestar.router.Router>`, subclasses of :class:`Controller <.controller.Controller>` or
            any function decorated by the route handler decorators.
        backend: The async backend to use, options are "asyncio" or "trio".
        backend_options: ``anyio`` options.
        base_url: URL scheme and domain for test request paths, e.g. ``http://testserver``.
        raise_server_exceptions: Flag for underlying the test client to raise server exceptions instead of wrapping them
            in an HTTP response.
        root_path: Path prefix for requests.
        session_config: Configuration for Session Middleware class to create raw session cookies for request to the
            route handlers.
        after_exception: A sequence of :class:`exception hook handlers <.types.AfterExceptionHookHandler>`. This
            hook is called after an exception occurs. In difference to exception handlers, it is not meant to
            return a response - only to process the exception (e.g. log it, send it to Sentry etc.).
        after_request: A sync or async function executed after the route handler function returned and the response
            object has been resolved. Receives the response object.
        after_response: A sync or async function called after the response has been awaited. It receives the
            :class:`Request <.connection.Request>` object and should not return any values.
        allowed_hosts: A sequence of allowed hosts, or an
            :class:`AllowedHostsConfig <.config.allowed_hosts.AllowedHostsConfig>` instance. Enables the builtin
            allowed hosts middleware.
        before_request: A sync or async function called immediately before calling the route handler. Receives the
            :class:`Request <.connection.Request>` instance and any non-``None`` return value is used for the
            response, bypassing the route handler.
        before_send: A sequence of :class:`before send hook handlers <.types.BeforeMessageSendHookHandler>`. Called
            when the ASGI send function is called.
        cache_control: A ``cache-control`` header of type
            :class:`CacheControlHeader <litestar.datastructures.CacheControlHeader>` to add to route handlers of
            this app. Can be overridden by route handlers.
        compression_config: Configures compression behaviour of the application, this enabled a builtin or user
            defined Compression middleware.
        cors_config: If set, configures :class:`CORSMiddleware <.middleware.cors.CORSMiddleware>`.
        csrf_config: If set, configures :class:`CSRFMiddleware <.middleware.csrf.CSRFMiddleware>`.
        debug: If ``True``, app errors rendered as HTML with a stack trace.
        dependencies: A string keyed mapping of dependency :class:`Providers <.di.Provide>`.
        dto: :class:`AbstractDTO <.dto.base_dto.AbstractDTO>` to use for (de)serializing and
            validation of request data.
        etag: An ``etag`` header of type :class:`ETag <.datastructures.ETag>` to add to route handlers of this app.
            Can be overridden by route handlers.
        event_emitter_backend: A subclass of
            :class:`BaseEventEmitterBackend <.events.emitter.BaseEventEmitterBackend>`.
        exception_handlers: A mapping of status codes and/or exception types to handler functions.
        guards: A sequence of :class:`Guard <.types.Guard>` callables.
        include_in_schema: A boolean flag dictating whether  the route handler should be documented in the OpenAPI schema.
        lifespan: A list of callables returning async context managers, wrapping the lifespan of the ASGI application
        listeners: A sequence of :class:`EventListener <.events.listener.EventListener>`.
        logging_config: A subclass of :class:`BaseLoggingConfig <.logging.config.BaseLoggingConfig>`.
        middleware: A sequence of :class:`Middleware <.types.Middleware>`.
        multipart_form_part_limit: The maximal number of allowed parts in a multipart/formdata request. This limit
            is intended to protect from DoS attacks.
        on_app_init: A sequence of :class:`OnAppInitHandler <.types.OnAppInitHandler>` instances. Handlers receive
            an instance of :class:`AppConfig <.config.app.AppConfig>` that will have been initially populated with
            the parameters passed to :class:`Litestar <litestar.app.Litestar>`, and must return an instance of same.
            If more than one handler is registered they are called in the order they are provided.
        on_shutdown: A sequence of :class:`LifespanHook <.types.LifespanHook>` called during application
            shutdown.
        on_startup: A sequence of :class:`LifespanHook <litestar.types.LifespanHook>` called during
            application startup.
        openapi_config: Defaults to :attr:`DEFAULT_OPENAPI_CONFIG`
        opt: A string keyed mapping of arbitrary values that can be accessed in :class:`Guards <.types.Guard>` or
            wherever you have access to :class:`Request <litestar.connection.request.Request>` or
            :class:`ASGI Scope <.types.Scope>`.
        parameters: A mapping of :class:`Parameter <.params.Parameter>` definitions available to all application
            paths.
        pdb_on_exception: Drop into the PDB when an exception occurs.
        plugins: Sequence of plugins.
        request_class: An optional subclass of :class:`Request <.connection.Request>` to use for http connections.
        response_class: A custom subclass of :class:`Response <.response.Response>` to be used as the app's default
            response.
        response_cookies: A sequence of :class:`Cookie <.datastructures.Cookie>`.
        response_headers: A string keyed mapping of :class:`ResponseHeader <.datastructures.ResponseHeader>`
        response_cache_config: Configures caching behavior of the application.
        return_dto: :class:`AbstractDTO <.dto.base_dto.AbstractDTO>` to use for serializing
            outbound response data.
        route_handlers: A sequence of route handlers, which can include instances of
            :class:`Router <.router.Router>`, subclasses of :class:`Controller <.controller.Controller>` or any
            callable decorated by the route handler decorators.
        security: A sequence of dicts that will be added to the schema of all route handlers in the application.
            See
            :data:`SecurityRequirement <.openapi.spec.SecurityRequirement>` for details.
        signature_namespace: A mapping of names to types for use in forward reference resolution during signature modeling.
        signature_types: A sequence of types for use in forward reference resolution during signature modeling.
            These types will be added to the signature namespace using their ``__name__`` attribute.
        state: An optional :class:`State <.datastructures.State>` for application state.
        static_files_config: A sequence of :class:`StaticFilesConfig <.static_files.StaticFilesConfig>`
        stores: Central registry of :class:`Store <.stores.base.Store>` that will be available throughout the
            application. If this is a dictionary to it will be passed to a
            :class:`StoreRegistry <.stores.registry.StoreRegistry>`. If it is a
            :class:`StoreRegistry <.stores.registry.StoreRegistry>`, this instance will be used directly.
        tags: A sequence of string tags that will be appended to the schema of all route handlers under the
            application.
        template_config: An instance of :class:`TemplateConfig <.template.TemplateConfig>`
        timeout: Request timeout
        type_encoders: A mapping of types to callables that transform them into types supported for serialization.
        websocket_class: An optional subclass of :class:`WebSocket <.connection.WebSocket>` to use for websocket
            connections.
        experimental_features: An iterable of experimental features to enable


    Returns:
        An instance of :class:`TestClient <.testing.TestClient>` with a created app instance.
    """
    route_handlers = () if route_handlers is None else route_handlers
    if is_class_and_subclass(route_handlers, Controller) or not isinstance(route_handlers, Sequence):
        route_handlers = (route_handlers,)

    app = Litestar(
        after_exception=after_exception,
        after_request=after_request,
        after_response=after_response,
        allowed_hosts=allowed_hosts,
        before_request=before_request,
        before_send=before_send,
        cache_control=cache_control,
        compression_config=compression_config,
        cors_config=cors_config,
        csrf_config=csrf_config,
        debug=debug,
        dependencies=dependencies,
        dto=dto,
        etag=etag,
        lifespan=lifespan,
        event_emitter_backend=event_emitter_backend,
        exception_handlers=exception_handlers,
        guards=guards,
        include_in_schema=include_in_schema,
        listeners=listeners,
        logging_config=logging_config,
        middleware=middleware,
        multipart_form_part_limit=multipart_form_part_limit,
        on_app_init=on_app_init,
        on_shutdown=on_shutdown,
        on_startup=on_startup,
        openapi_config=openapi_config,
        opt=opt,
        parameters=parameters,
        pdb_on_exception=pdb_on_exception,
        plugins=plugins,
        request_class=request_class,
        response_cache_config=response_cache_config,
        response_class=response_class,
        response_cookies=response_cookies,
        response_headers=response_headers,
        return_dto=return_dto,
        route_handlers=route_handlers,
        security=security,
        signature_namespace=signature_namespace,
        signature_types=signature_types,
        state=state,
        static_files_config=static_files_config,
        stores=stores,
        tags=tags,
        template_config=template_config,
        type_encoders=type_encoders,
        websocket_class=websocket_class,
        experimental_features=experimental_features,
    )

    return TestClient[Litestar](
        app=app,
        backend=backend,
        backend_options=backend_options,
        base_url=base_url,
        raise_server_exceptions=raise_server_exceptions,
        root_path=root_path,
        session_config=session_config,
        timeout=timeout,
    )


def create_async_test_client(
    route_handlers: ControllerRouterHandler | Sequence[ControllerRouterHandler] | None = None,
    *,
    after_exception: Sequence[AfterExceptionHookHandler] | None = None,
    after_request: AfterRequestHookHandler | None = None,
    after_response: AfterResponseHookHandler | None = None,
    allowed_hosts: Sequence[str] | AllowedHostsConfig | None = None,
    backend: Literal["asyncio", "trio"] = "asyncio",
    backend_options: Mapping[str, Any] | None = None,
    base_url: str = "http://testserver.local",
    before_request: BeforeRequestHookHandler | None = None,
    before_send: Sequence[BeforeMessageSendHookHandler] | None = None,
    cache_control: CacheControlHeader | None = None,
    compression_config: CompressionConfig | None = None,
    cors_config: CORSConfig | None = None,
    csrf_config: CSRFConfig | None = None,
    debug: bool = True,
    dependencies: Dependencies | None = None,
    dto: type[AbstractDTO] | None | EmptyType = Empty,
    etag: ETag | None = None,
    event_emitter_backend: type[BaseEventEmitterBackend] = SimpleEventEmitter,
    exception_handlers: ExceptionHandlersMap | None = None,
    guards: Sequence[Guard] | None = None,
    include_in_schema: bool | EmptyType = Empty,
    lifespan: list[Callable[[Litestar], AbstractAsyncContextManager] | AbstractAsyncContextManager] | None = None,
    listeners: Sequence[EventListener] | None = None,
    logging_config: BaseLoggingConfig | EmptyType | None = Empty,
    middleware: Sequence[Middleware] | None = None,
    multipart_form_part_limit: int = 1000,
    on_app_init: Sequence[OnAppInitHandler] | None = None,
    on_shutdown: Sequence[LifespanHook] | None = None,
    on_startup: Sequence[LifespanHook] | None = None,
    openapi_config: OpenAPIConfig | None = DEFAULT_OPENAPI_CONFIG,
    opt: Mapping[str, Any] | None = None,
    parameters: ParametersMap | None = None,
    pdb_on_exception: bool | None = None,
    plugins: Sequence[PluginProtocol] | None = None,
    raise_server_exceptions: bool = True,
    request_class: type[Request] | None = None,
    response_cache_config: ResponseCacheConfig | None = None,
    response_class: ResponseType | None = None,
    response_cookies: ResponseCookies | None = None,
    response_headers: Sequence[ResponseHeader] | None = None,
    return_dto: type[AbstractDTO] | None | EmptyType = Empty,
    root_path: str = "",
    security: Sequence[SecurityRequirement] | None = None,
    session_config: BaseBackendConfig | None = None,
    signature_namespace: Mapping[str, Any] | None = None,
    signature_types: Sequence[Any] | None = None,
    state: State | None = None,
    static_files_config: Sequence[StaticFilesConfig] | None = None,
    stores: StoreRegistry | dict[str, Store] | None = None,
    tags: Sequence[str] | None = None,
    template_config: TemplateConfig | None = None,
    timeout: float | None = None,
    type_encoders: TypeEncodersMap | None = None,
    websocket_class: type[WebSocket] | None = None,
    experimental_features: list[ExperimentalFeatures] | None = None,
) -> AsyncTestClient[Litestar]:
    """Create a Litestar app instance and initializes it.

    :class:`TestClient <litestar.testing.TestClient>` with it.

    Notes:
        - This function should be called as a context manager to ensure async startup and shutdown are
            handled correctly.

    Examples:
        .. code-block:: python

            from litestar import get
            from litestar.testing import create_test_client


            @get("/some-path")
            def my_handler() -> dict[str, str]:
                return {"hello": "world"}


            def test_my_handler() -> None:
                with create_test_client(my_handler) as client:
                    response == client.get("/some-path")
                    assert response.json() == {"hello": "world"}

    Args:
        route_handlers: A single handler or a sequence of route handlers, which can include instances of
            :class:`Router <litestar.router.Router>`, subclasses of :class:`Controller <.controller.Controller>` or
            any function decorated by the route handler decorators.
        backend: The async backend to use, options are "asyncio" or "trio".
        backend_options: ``anyio`` options.
        base_url: URL scheme and domain for test request paths, e.g. ``http://testserver``.
        raise_server_exceptions: Flag for underlying the test client to raise server exceptions instead of wrapping them
            in an HTTP response.
        root_path: Path prefix for requests.
        session_config: Configuration for Session Middleware class to create raw session cookies for request to the
            route handlers.
        after_exception: A sequence of :class:`exception hook handlers <.types.AfterExceptionHookHandler>`. This
            hook is called after an exception occurs. In difference to exception handlers, it is not meant to
            return a response - only to process the exception (e.g. log it, send it to Sentry etc.).
        after_request: A sync or async function executed after the route handler function returned and the response
            object has been resolved. Receives the response object.
        after_response: A sync or async function called after the response has been awaited. It receives the
            :class:`Request <.connection.Request>` object and should not return any values.
        allowed_hosts: A sequence of allowed hosts, or an
            :class:`AllowedHostsConfig <.config.allowed_hosts.AllowedHostsConfig>` instance. Enables the builtin
            allowed hosts middleware.
        before_request: A sync or async function called immediately before calling the route handler. Receives the
            :class:`Request <.connection.Request>` instance and any non-``None`` return value is used for the
            response, bypassing the route handler.
        before_send: A sequence of :class:`before send hook handlers <.types.BeforeMessageSendHookHandler>`. Called
            when the ASGI send function is called.
        cache_control: A ``cache-control`` header of type
            :class:`CacheControlHeader <litestar.datastructures.CacheControlHeader>` to add to route handlers of
            this app. Can be overridden by route handlers.
        compression_config: Configures compression behaviour of the application, this enabled a builtin or user
            defined Compression middleware.
        cors_config: If set, configures :class:`CORSMiddleware <.middleware.cors.CORSMiddleware>`.
        csrf_config: If set, configures :class:`CSRFMiddleware <.middleware.csrf.CSRFMiddleware>`.
        debug: If ``True``, app errors rendered as HTML with a stack trace.
        dependencies: A string keyed mapping of dependency :class:`Providers <.di.Provide>`.
        dto: :class:`AbstractDTO <.dto.base_dto.AbstractDTO>` to use for (de)serializing and
            validation of request data.
        etag: An ``etag`` header of type :class:`ETag <.datastructures.ETag>` to add to route handlers of this app.
            Can be overridden by route handlers.
        event_emitter_backend: A subclass of
            :class:`BaseEventEmitterBackend <.events.emitter.BaseEventEmitterBackend>`.
        exception_handlers: A mapping of status codes and/or exception types to handler functions.
        guards: A sequence of :class:`Guard <.types.Guard>` callables.
        include_in_schema: A boolean flag dictating whether  the route handler should be documented in the OpenAPI schema.
        lifespan: A list of callables returning async context managers, wrapping the lifespan of the ASGI application
        listeners: A sequence of :class:`EventListener <.events.listener.EventListener>`.
        logging_config: A subclass of :class:`BaseLoggingConfig <.logging.config.BaseLoggingConfig>`.
        middleware: A sequence of :class:`Middleware <.types.Middleware>`.
        multipart_form_part_limit: The maximal number of allowed parts in a multipart/formdata request. This limit
            is intended to protect from DoS attacks.
        on_app_init: A sequence of :class:`OnAppInitHandler <.types.OnAppInitHandler>` instances. Handlers receive
            an instance of :class:`AppConfig <.config.app.AppConfig>` that will have been initially populated with
            the parameters passed to :class:`Litestar <litestar.app.Litestar>`, and must return an instance of same.
            If more than one handler is registered they are called in the order they are provided.
        on_shutdown: A sequence of :class:`LifespanHook <.types.LifespanHook>` called during application
            shutdown.
        on_startup: A sequence of :class:`LifespanHook <litestar.types.LifespanHook>` called during
            application startup.
        openapi_config: Defaults to :attr:`DEFAULT_OPENAPI_CONFIG`
        opt: A string keyed mapping of arbitrary values that can be accessed in :class:`Guards <.types.Guard>` or
            wherever you have access to :class:`Request <litestar.connection.request.Request>` or
            :class:`ASGI Scope <.types.Scope>`.
        parameters: A mapping of :class:`Parameter <.params.Parameter>` definitions available to all application
            paths.
        pdb_on_exception: Drop into the PDB when an exception occurs.
        plugins: Sequence of plugins.
        request_class: An optional subclass of :class:`Request <.connection.Request>` to use for http connections.
        response_class: A custom subclass of :class:`Response <.response.Response>` to be used as the app's default
            response.
        response_cookies: A sequence of :class:`Cookie <.datastructures.Cookie>`.
        response_headers: A string keyed mapping of :class:`ResponseHeader <.datastructures.ResponseHeader>`
        response_cache_config: Configures caching behavior of the application.
        return_dto: :class:`AbstractDTO <.dto.base_dto.AbstractDTO>` to use for serializing
            outbound response data.
        route_handlers: A sequence of route handlers, which can include instances of
            :class:`Router <.router.Router>`, subclasses of :class:`Controller <.controller.Controller>` or any
            callable decorated by the route handler decorators.
        security: A sequence of dicts that will be added to the schema of all route handlers in the application.
            See
            :data:`SecurityRequirement <.openapi.spec.SecurityRequirement>` for details.
        signature_namespace: A mapping of names to types for use in forward reference resolution during signature modeling.
        signature_types: A sequence of types for use in forward reference resolution during signature modeling.
            These types will be added to the signature namespace using their ``__name__`` attribute.
        state: An optional :class:`State <.datastructures.State>` for application state.
        static_files_config: A sequence of :class:`StaticFilesConfig <.static_files.StaticFilesConfig>`
        stores: Central registry of :class:`Store <.stores.base.Store>` that will be available throughout the
            application. If this is a dictionary to it will be passed to a
            :class:`StoreRegistry <.stores.registry.StoreRegistry>`. If it is a
            :class:`StoreRegistry <.stores.registry.StoreRegistry>`, this instance will be used directly.
        tags: A sequence of string tags that will be appended to the schema of all route handlers under the
            application.
        template_config: An instance of :class:`TemplateConfig <.template.TemplateConfig>`
        timeout: Request timeout
        type_encoders: A mapping of types to callables that transform them into types supported for serialization.
        websocket_class: An optional subclass of :class:`WebSocket <.connection.WebSocket>` to use for websocket
            connections.
        experimental_features: An iterable of experimental features to enable

    Returns:
        An instance of :class:`AsyncTestClient <litestar.testing.AsyncTestClient>` with a created app instance.
    """
    route_handlers = () if route_handlers is None else route_handlers
    if is_class_and_subclass(route_handlers, Controller) or not isinstance(route_handlers, Sequence):
        route_handlers = (route_handlers,)

    app = Litestar(
        after_exception=after_exception,
        after_request=after_request,
        after_response=after_response,
        allowed_hosts=allowed_hosts,
        before_request=before_request,
        before_send=before_send,
        cache_control=cache_control,
        compression_config=compression_config,
        cors_config=cors_config,
        csrf_config=csrf_config,
        debug=debug,
        dependencies=dependencies,
        dto=dto,
        etag=etag,
        event_emitter_backend=event_emitter_backend,
        exception_handlers=exception_handlers,
        guards=guards,
        include_in_schema=include_in_schema,
        lifespan=lifespan,
        listeners=listeners,
        logging_config=logging_config,
        middleware=middleware,
        multipart_form_part_limit=multipart_form_part_limit,
        on_app_init=on_app_init,
        on_shutdown=on_shutdown,
        on_startup=on_startup,
        openapi_config=openapi_config,
        opt=opt,
        parameters=parameters,
        pdb_on_exception=pdb_on_exception,
        plugins=plugins,
        request_class=request_class,
        response_cache_config=response_cache_config,
        response_class=response_class,
        response_cookies=response_cookies,
        response_headers=response_headers,
        return_dto=return_dto,
        route_handlers=route_handlers,
        security=security,
        signature_namespace=signature_namespace,
        signature_types=signature_types,
        state=state,
        static_files_config=static_files_config,
        stores=stores,
        tags=tags,
        template_config=template_config,
        type_encoders=type_encoders,
        websocket_class=websocket_class,
        experimental_features=experimental_features,
    )

    return AsyncTestClient[Litestar](
        app=app,
        backend=backend,
        backend_options=backend_options,
        base_url=base_url,
        raise_server_exceptions=raise_server_exceptions,
        root_path=root_path,
        session_config=session_config,
        timeout=timeout,
    )



from __future__ import annotations

import math
import sys
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import AsyncExitStack
from functools import partial
from typing import TYPE_CHECKING, Any, Sequence

if sys.version_info < (3, 9):
    from typing import AsyncContextManager
else:
    from contextlib import AbstractAsyncContextManager as AsyncContextManager

import anyio

from litestar.exceptions import ImproperlyConfiguredException

if TYPE_CHECKING:
    from types import TracebackType

    from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream

    from litestar.events.listener import EventListener

__all__ = ("BaseEventEmitterBackend", "SimpleEventEmitter")


class BaseEventEmitterBackend(AsyncContextManager["BaseEventEmitterBackend"], ABC):
    """Abstract class used to define event emitter backends."""

    __slots__ = ("listeners",)

    listeners: defaultdict[str, set[EventListener]]

    def __init__(self, listeners: Sequence[EventListener]) -> None:
        """Create an event emitter instance.

        Args:
            listeners: A list of listeners.
        """
        self.listeners = defaultdict(set)
        for listener in listeners:
            for event_id in listener.event_ids:
                self.listeners[event_id].add(listener)

    @abstractmethod
    def emit(self, event_id: str, *args: Any, **kwargs: Any) -> None:
        """Emit an event to all attached listeners.

        Args:
            event_id: The ID of the event to emit, e.g 'my_event'.
            *args: args to pass to the listener(s).
            **kwargs: kwargs to pass to the listener(s)

        Returns:
            None
        """
        raise NotImplementedError("not implemented")


class SimpleEventEmitter(BaseEventEmitterBackend):
    """Event emitter the works only in the current process"""

    __slots__ = ("_queue", "_exit_stack", "_receive_stream", "_send_stream")

    def __init__(self, listeners: Sequence[EventListener]) -> None:
        """Create an event emitter instance.

        Args:
            listeners: A list of listeners.
        """
        super().__init__(listeners=listeners)
        self._receive_stream: MemoryObjectReceiveStream | None = None
        self._send_stream: MemoryObjectSendStream | None = None
        self._exit_stack: AsyncExitStack | None = None

    async def _worker(self, receive_stream: MemoryObjectReceiveStream) -> None:
        """Run items from ``receive_stream`` in a task group.

        Returns:
            None
        """
        async with receive_stream, anyio.create_task_group() as task_group:
            async for item in receive_stream:
                fn, args, kwargs = item
                if kwargs:
                    fn = partial(fn, **kwargs)
                task_group.start_soon(fn, *args)

    async def __aenter__(self) -> SimpleEventEmitter:
        self._exit_stack = AsyncExitStack()
        send_stream, receive_stream = anyio.create_memory_object_stream(math.inf)  # type: ignore[var-annotated]
        self._send_stream = send_stream
        task_group = anyio.create_task_group()

        await self._exit_stack.enter_async_context(task_group)
        await self._exit_stack.enter_async_context(send_stream)
        task_group.start_soon(self._worker, receive_stream)

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self._exit_stack:
            await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)

        self._exit_stack = None
        self._send_stream = None

    def emit(self, event_id: str, *args: Any, **kwargs: Any) -> None:
        """Emit an event to all attached listeners.

        Args:
            event_id: The ID of the event to emit, e.g 'my_event'.
            *args: args to pass to the listener(s).
            **kwargs: kwargs to pass to the listener(s)

        Returns:
            None
        """
        if not (self._send_stream and self._exit_stack):
            raise RuntimeError("Emitter not initialized")

        if listeners := self.listeners.get(event_id):
            for listener in listeners:
                self._send_stream.send_nowait((listener.fn, args, kwargs))
            return
        raise ImproperlyConfiguredException(f"no event listeners are registered for event ID: {event_id}")



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Callable, Mapping, Sequence

from litestar.contrib.prometheus.middleware import (
    PrometheusMiddleware,
)
from litestar.exceptions import MissingDependencyException
from litestar.middleware.base import DefineMiddleware

__all__ = ("PrometheusConfig",)


try:
    import prometheus_client  # noqa: F401
except ImportError as e:
    raise MissingDependencyException("prometheus_client") from e


if TYPE_CHECKING:
    from litestar.connection.request import Request
    from litestar.types import Method, Scopes


@dataclass
class PrometheusConfig:
    """Configuration class for the PrometheusConfig middleware."""

    app_name: str = field(default="litestar")
    """The name of the application to use in the metrics."""
    prefix: str = "litestar"
    """The prefix to use for the metrics."""
    labels: Mapping[str, str | Callable] | None = field(default=None)
    """A mapping of labels to add to the metrics. The values can be either a string or a callable that returns a string."""
    exemplars: Callable[[Request], dict] | None = field(default=None)
    """A callable that returns a list of exemplars to add to the metrics. Only supported in opementrics-text exposition format."""
    buckets: list[str | float] | None = field(default=None)
    """A list of buckets to use for the histogram."""
    excluded_http_methods: Method | Sequence[Method] | None = field(default=None)
    """A list of http methods to exclude from the metrics."""
    exclude_unhandled_paths: bool = field(default=False)
    """Whether to ignore requests for unhandled paths from the metrics."""
    exclude: str | list[str] | None = field(default=None)
    """A pattern or list of patterns for routes to exclude from the metrics."""
    exclude_opt_key: str | None = field(default=None)
    """A key or list of keys in ``opt`` with which a route handler can "opt-out" of the middleware."""
    scopes: Scopes | None = field(default=None)
    """ASGI scopes processed by the middleware, if None both ``http`` and ``websocket`` will be processed."""
    middleware_class: type[PrometheusMiddleware] = field(default=PrometheusMiddleware)
    """The middleware class to use.
    """

    @property
    def middleware(self) -> DefineMiddleware:
        """Create an instance of :class:`DefineMiddleware <litestar.middleware.base.DefineMiddleware>` that wraps with.

        [PrometheusMiddleware][litestar.contrib.prometheus.PrometheusMiddleware]. or a subclass
        of this middleware.

        Returns:
            An instance of ``DefineMiddleware``.
        """
        return DefineMiddleware(self.middleware_class, config=self)



from __future__ import annotations

from typing import TYPE_CHECKING

import pydantic as pydantic_v2
from pydantic import v1 as pydantic_v1
from typing_extensions import Annotated

from litestar.contrib.pydantic import PydanticDTO
from litestar.dto import dto_field
from litestar.dto.data_structures import DTOFieldDefinition
from litestar.typing import FieldDefinition

if TYPE_CHECKING:
    from typing import Callable


def test_field_definition_generation_v1(
    int_factory: Callable[[], int],
    expected_field_defs: list[DTOFieldDefinition],
) -> None:
    class TestModel(pydantic_v1.BaseModel):
        a: int
        b: Annotated[int, pydantic_v1.Field(**dto_field("read-only"))]  # pyright: ignore
        c: Annotated[int, pydantic_v1.Field(gt=1)]
        d: int = pydantic_v1.Field(default=1)
        e: int = pydantic_v1.Field(default_factory=int_factory)

    field_defs = list(PydanticDTO.generate_field_definitions(TestModel))
    assert field_defs[0].model_name == "TestModel"
    for field_def, exp in zip(field_defs, expected_field_defs):
        assert field_def == exp


def test_field_definition_generation_v2(
    int_factory: Callable[[], int],
    expected_field_defs: list[DTOFieldDefinition],
) -> None:
    class TestModel(pydantic_v2.BaseModel):
        a: int
        b: Annotated[int, pydantic_v2.Field(**dto_field("read-only"))]  # pyright: ignore
        c: Annotated[int, pydantic_v2.Field(gt=1)]
        d: int = pydantic_v2.Field(default=1)
        e: int = pydantic_v2.Field(default_factory=int_factory)

    field_defs = list(PydanticDTO.generate_field_definitions(TestModel))
    assert field_defs[0].model_name == "TestModel"
    for field_def, exp in zip(field_defs, expected_field_defs):
        assert field_def == exp


def test_detect_nested_field(base_model: type[pydantic_v1.BaseModel | pydantic_v2.BaseModel]) -> None:
    class TestModel(base_model):  # type: ignore[misc, valid-type]
        a: int

    class NotModel:
        pass

    assert PydanticDTO.detect_nested_field(FieldDefinition.from_annotation(TestModel)) is True
    assert PydanticDTO.detect_nested_field(FieldDefinition.from_annotation(NotModel)) is False



import sys
from types import ModuleType
from typing import Any, Callable

import pytest
from msgspec import Struct

from litestar import post
from litestar.testing import create_test_client
from tests.models import DataclassPerson, MsgSpecStructPerson, TypedDictPerson


@pytest.mark.parametrize("cls", (DataclassPerson, TypedDictPerson, MsgSpecStructPerson))
def test_spec_generation(cls: Any) -> None:
    @post("/")
    def handler(data: cls) -> cls:
        return data

    with create_test_client(handler) as client:
        schema = client.app.openapi_schema
        assert schema
        assert schema.to_schema()["components"]["schemas"][cls.__name__] == {
            "properties": {
                "first_name": {"type": "string"},
                "last_name": {"type": "string"},
                "id": {"type": "string"},
                "optional": {"oneOf": [{"type": "null"}, {"type": "string"}]},
                "complex": {
                    "type": "object",
                    "additionalProperties": {
                        "type": "array",
                        "items": {"type": "object", "additionalProperties": {"type": "string"}},
                    },
                },
                "pets": {
                    "oneOf": [
                        {"type": "null"},
                        {
                            "items": {"$ref": "#/components/schemas/DataclassPet"},
                            "type": "array",
                        },
                    ]
                },
            },
            "type": "object",
            "required": ["complex", "first_name", "id", "last_name"],
            "title": f"{cls.__name__}",
        }


def test_msgspec_schema() -> None:
    class CamelizedStruct(Struct, rename="camel"):
        field_one: int
        field_two: float

    @post("/")
    def handler(data: CamelizedStruct) -> CamelizedStruct:
        return data

    with create_test_client(handler) as client:
        schema = client.app.openapi_schema
        assert schema

        assert schema.to_schema()["components"]["schemas"]["test_msgspec_schema.CamelizedStruct"] == {
            "properties": {"fieldOne": {"type": "integer"}, "fieldTwo": {"type": "number"}},
            "required": ["fieldOne", "fieldTwo"],
            "title": "CamelizedStruct",
            "type": "object",
        }


@pytest.fixture()
def py_38_module_content() -> str:
    return """
from __future__ import annotations

from typing import List, Optional

from msgspec import Struct

from litestar import Litestar, get

class A(Struct):
    a: A
    b: B
    opt_a: Optional[A] = None
    opt_b: Optional[B] = None
    list_a: List[A] = []
    list_b: List[B] = []

class B(Struct):
    a: A
    b: B
    opt_a: Optional[A] = None
    opt_b: Optional[B] = None
    list_a: List[A] = []
    list_b: List[B] = []

@get("/")
async def test() -> A:
    return A()
"""


@pytest.fixture()
def py_310_module_content() -> str:
    return """
from __future__ import annotations

from msgspec import Struct

from litestar import Litestar, get

class A(Struct):
    a: A
    b: B
    opt_a: A | None = None
    opt_b: B | None = None
    list_a: list[A] = []
    list_b: list[B] = []

class B(Struct):
    a: A
    b: B
    opt_a: A | None = None
    opt_b: B | None = None
    list_a: list[A] = []
    list_b: list[B] = []

@get("/")
async def test() -> A:
    return A()
"""


@pytest.mark.parametrize(
    ("fixture_name",),
    [
        ("py_38_module_content",),
        pytest.param(
            "py_310_module_content",
            marks=pytest.mark.skipif(
                sys.version_info < (3, 10),
                reason="requires python 3.10",
            ),
        ),
    ],
)
def test_recursive_schema_generation(
    fixture_name: str, create_module: Callable[[str], ModuleType], request: pytest.FixtureRequest
) -> None:
    module_content = request.getfixturevalue(fixture_name)
    module = create_module(module_content)
    with create_test_client(module.test, debug=True) as client:
        schema = client.app.openapi_schema
        assert schema
        assert schema.to_schema()["components"]["schemas"]["A"] == {
            "required": ["a", "b"],
            "properties": {
                "a": {"$ref": "#/components/schemas/A"},
                "b": {"$ref": "#/components/schemas/B"},
                "opt_a": {"oneOf": [{"type": "null"}, {"$ref": "#/components/schemas/A"}]},
                "opt_b": {"oneOf": [{"type": "null"}, {"$ref": "#/components/schemas/B"}]},
                "list_a": {"items": {"$ref": "#/components/schemas/A"}, "type": "array"},
                "list_b": {"items": {"$ref": "#/components/schemas/B"}, "type": "array"},
            },
            "type": "object",
            "title": "A",
        }
        assert schema.to_schema()["components"]["schemas"]["B"] == {
            "required": ["a", "b"],
            "properties": {
                "a": {"$ref": "#/components/schemas/A"},
                "b": {"$ref": "#/components/schemas/B"},
                "opt_a": {"oneOf": [{"type": "null"}, {"$ref": "#/components/schemas/A"}]},
                "opt_b": {"oneOf": [{"type": "null"}, {"$ref": "#/components/schemas/B"}]},
                "list_a": {"items": {"$ref": "#/components/schemas/A"}, "type": "array"},
                "list_b": {"items": {"$ref": "#/components/schemas/B"}, "type": "array"},
            },
            "type": "object",
            "title": "B",
        }



from docs.examples.middleware.call_order import app

from litestar.testing import TestClient


def test_call_order() -> None:
    with TestClient(app=app) as client:
        response = client.get("/router/controller/handler")
        assert response.status_code == 200
        assert response.json() == [0, 1, 2, 3, 4, 5, 6, 7]



from __future__ import annotations

from dataclasses import dataclass

from litestar.openapi.spec.base import BaseSchemaObject

__all__ = ("Contact",)


@dataclass
class Contact(BaseSchemaObject):
    """Contact information for the exposed API."""

    name: str | None = None
    """The identifying name of the contact person/organization."""

    url: str | None = None
    """The URL pointing to the contact information. MUST be in the form of a URL."""

    email: str | None = None
    """The email address of the contact person/organization. MUST be in the form of an email address."""



from dataclasses import asdict

from litestar import post
from litestar.params import Body
from litestar.status_codes import HTTP_201_CREATED
from litestar.testing import create_test_client

from . import Form


def test_request_body_json() -> None:
    @post(path="/test")
    def test_method(data: Form = Body()) -> None:
        assert isinstance(data, Form)

    with create_test_client(test_method) as client:
        response = client.post("/test", json=asdict(Form(name="Moishe Zuchmir", age=30, programmer=True, value="100")))
        assert response.status_code == HTTP_201_CREATED


def test_empty_dict_allowed() -> None:
    @post(path="/test")
    def test_method(data: dict) -> None:
        assert isinstance(data, dict)

    with create_test_client(test_method) as client:
        response = client.post("/test", json={})
        assert response.status_code == HTTP_201_CREATED



# ruff: noqa: UP006, UP007
from __future__ import annotations

import re
from functools import partial
from pathlib import Path, PurePath
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Literal,
    Optional,
    Sequence,
    Set,
    Type,
    TypedDict,
    Union,
    cast,
)
from uuid import UUID

from msgspec import NODEFAULT, Meta, Struct, ValidationError, convert, defstruct
from msgspec.structs import asdict
from typing_extensions import Annotated

from litestar._signature.types import ExtendedMsgSpecValidationError
from litestar._signature.utils import (
    _get_decoder_for_type,
    _normalize_annotation,
    _validate_signature_dependencies,
)
from litestar.datastructures.state import ImmutableState
from litestar.dto import AbstractDTO, DTOData
from litestar.enums import ParamType, ScopeType
from litestar.exceptions import InternalServerException, ValidationException
from litestar.params import KwargDefinition, ParameterKwarg
from litestar.typing import FieldDefinition  # noqa
from litestar.utils import is_class_and_subclass
from litestar.utils.dataclass import simple_asdict

if TYPE_CHECKING:
    from typing_extensions import NotRequired

    from litestar.connection import ASGIConnection
    from litestar.types import AnyCallable, TypeDecodersSequence
    from litestar.utils.signature import ParsedSignature

__all__ = (
    "ErrorMessage",
    "SignatureModel",
)


class ErrorMessage(TypedDict):
    # key may not be set in some cases, like when a query param is set but
    # doesn't match the required length during `attrs` validation
    # in this case, we don't show a key at all as it will be empty
    key: NotRequired[str]
    message: str
    source: NotRequired[Literal["body"] | ParamType]


MSGSPEC_CONSTRAINT_FIELDS = (
    "gt",
    "ge",
    "lt",
    "le",
    "multiple_of",
    "pattern",
    "min_length",
    "max_length",
)

ERR_RE = re.compile(r"`\$\.(.+)`$")

DEFAULT_TYPE_DECODERS = [
    (lambda x: is_class_and_subclass(x, (Path, PurePath, ImmutableState, UUID)), lambda t, v: t(v)),
]


def _deserializer(target_type: Any, value: Any, default_deserializer: Callable[[Any, Any], Any]) -> Any:
    if isinstance(value, DTOData):
        return value

    if isinstance(value, target_type):
        return value

    if decoder := getattr(target_type, "_decoder", None):
        return decoder(target_type, value)

    return default_deserializer(target_type, value)


class SignatureModel(Struct):
    """Model that represents a function signature that uses a msgspec specific type or types."""

    _data_dto: ClassVar[Optional[Type[AbstractDTO]]]
    _dependency_name_set: ClassVar[Set[str]]
    _fields: ClassVar[Dict[str, FieldDefinition]]
    _return_annotation: ClassVar[Any]

    @classmethod
    def _create_exception(cls, connection: ASGIConnection, messages: list[ErrorMessage]) -> Exception:
        """Create an exception class - either a ValidationException or an InternalServerException, depending on whether
            the failure is in client provided values or injected dependencies.

        Args:
            connection: An ASGI connection instance.
            messages: A list of error messages.

        Returns:
            An Exception
        """
        method = connection.method if hasattr(connection, "method") else ScopeType.WEBSOCKET  # pyright: ignore
        if client_errors := [
            err_message
            for err_message in messages
            if ("key" in err_message and err_message["key"] not in cls._dependency_name_set) or "key" not in err_message
        ]:
            return ValidationException(detail=f"Validation failed for {method} {connection.url}", extra=client_errors)
        return InternalServerException()

    @classmethod
    def _build_error_message(cls, keys: Sequence[str], exc_msg: str, connection: ASGIConnection) -> ErrorMessage:
        """Build an error message.

        Args:
            keys: A list of keys.
            exc_msg: A message.
            connection: An ASGI connection instance.

        Returns:
            An ErrorMessage
        """

        message: ErrorMessage = {"message": exc_msg.split(" - ")[0]}

        if keys:
            message["key"] = key = ".".join(keys)
            if keys[0].startswith("data"):
                message["key"] = message["key"].replace("data.", "")
                message["source"] = "body"
            elif key in connection.query_params:
                message["source"] = ParamType.QUERY

            elif key in cls._fields and isinstance(cls._fields[key].kwarg_definition, ParameterKwarg):
                if cast(ParameterKwarg, cls._fields[key].kwarg_definition).cookie:
                    message["source"] = ParamType.COOKIE
                elif cast(ParameterKwarg, cls._fields[key].kwarg_definition).header:
                    message["source"] = ParamType.HEADER
                else:
                    message["source"] = ParamType.QUERY

        return message

    @classmethod
    def _collect_errors(cls, deserializer: Callable[[Any, Any], Any], **kwargs: Any) -> list[tuple[str, Exception]]:
        exceptions: list[tuple[str, Exception]] = []
        for field_name in cls._fields:
            try:
                raw_value = kwargs[field_name]
                annotation = cls.__annotations__[field_name]
                convert(raw_value, type=annotation, strict=False, dec_hook=deserializer, str_keys=True)
            except Exception as e:  # noqa: BLE001
                exceptions.append((field_name, e))

        return exceptions

    @classmethod
    def parse_values_from_connection_kwargs(cls, connection: ASGIConnection, **kwargs: Any) -> dict[str, Any]:
        """Extract values from the connection instance and return a dict of parsed values.

        Args:
            connection: The ASGI connection instance.
            **kwargs: A dictionary of kwargs.

        Raises:
            ValidationException: If validation failed.
            InternalServerException: If another exception has been raised.

        Returns:
            A dictionary of parsed values
        """
        messages: list[ErrorMessage] = []
        deserializer = partial(_deserializer, default_deserializer=connection.route_handler.default_deserializer)
        try:
            return convert(kwargs, cls, strict=False, dec_hook=deserializer, str_keys=True).to_dict()
        except ExtendedMsgSpecValidationError as e:
            for exc in e.errors:
                keys = [str(loc) for loc in exc["loc"]]
                message = cls._build_error_message(keys=keys, exc_msg=exc["msg"], connection=connection)
                messages.append(message)
            raise cls._create_exception(messages=messages, connection=connection) from e
        except ValidationError as e:
            for field_name, exc in cls._collect_errors(deserializer=deserializer, **kwargs):  # type: ignore[assignment]
                match = ERR_RE.search(str(exc))
                keys = [field_name, str(match.group(1))] if match else [field_name]
                message = cls._build_error_message(keys=keys, exc_msg=str(exc), connection=connection)
                messages.append(message)
            raise cls._create_exception(messages=messages, connection=connection) from e

    def to_dict(self) -> dict[str, Any]:
        """Normalize access to the signature model's dictionary method, because different backends use different methods
        for this.

        Returns: A dictionary of string keyed values.
        """
        return asdict(self)

    @classmethod
    def create(
        cls,
        dependency_name_set: set[str],
        fn: AnyCallable,
        parsed_signature: ParsedSignature,
        type_decoders: TypeDecodersSequence,
        data_dto: type[AbstractDTO] | None = None,
    ) -> type[SignatureModel]:
        fn_name = (
            fn_name if (fn_name := getattr(fn, "__name__", "anonymous")) and fn_name != "<lambda>" else "anonymous"
        )

        dependency_names = _validate_signature_dependencies(
            dependency_name_set=dependency_name_set, fn_name=fn_name, parsed_signature=parsed_signature
        )

        struct_fields: list[tuple[str, Any, Any]] = []

        for field_definition in parsed_signature.parameters.values():
            meta_data: Meta | None = None

            if isinstance(field_definition.kwarg_definition, KwargDefinition):
                meta_kwargs: dict[str, Any] = {"extra": {}}

                kwarg_definition = simple_asdict(field_definition.kwarg_definition, exclude_empty=True)
                if min_items := kwarg_definition.pop("min_items", None):
                    meta_kwargs["min_length"] = min_items
                if max_items := kwarg_definition.pop("max_items", None):
                    meta_kwargs["max_length"] = max_items

                for k, v in kwarg_definition.items():
                    if hasattr(Meta, k) and v is not None:
                        meta_kwargs[k] = v
                    else:
                        meta_kwargs["extra"][k] = v

                meta_data = Meta(**meta_kwargs)

            annotation = cls._create_annotation(
                field_definition=field_definition,
                type_decoders=[*(type_decoders or []), *DEFAULT_TYPE_DECODERS],
                meta_data=meta_data,
                data_dto=data_dto,
            )

            default = field_definition.default if field_definition.has_default else NODEFAULT
            struct_fields.append((field_definition.name, annotation, default))

        return defstruct(  # type:ignore[return-value]
            f"{fn_name}_signature_model",
            struct_fields,
            bases=(cls,),
            module=getattr(fn, "__module__", None),
            namespace={
                "_return_annotation": parsed_signature.return_type.annotation,
                "_dependency_name_set": dependency_names,
                "_fields": parsed_signature.parameters,
                "_data_dto": data_dto,
            },
            kw_only=True,
        )

    @classmethod
    def _create_annotation(
        cls,
        field_definition: FieldDefinition,
        type_decoders: TypeDecodersSequence,
        meta_data: Meta | None = None,
        data_dto: type[AbstractDTO] | None = None,
    ) -> Any:
        # DTOs have already validated their data, so we can just use Any here
        if field_definition.name == "data" and data_dto:
            return Any

        annotation = _normalize_annotation(field_definition=field_definition)

        if annotation is Any:
            return annotation

        if field_definition.is_union:
            types = [
                cls._create_annotation(
                    field_definition=inner_type,
                    type_decoders=type_decoders,
                    meta_data=meta_data,
                )
                for inner_type in field_definition.inner_types
                if not inner_type.is_none_type
            ]
            return Optional[Union[tuple(types)]] if field_definition.is_optional else Union[tuple(types)]  # pyright: ignore

        if decoder := _get_decoder_for_type(annotation, type_decoders=type_decoders):
            # FIXME: temporary (hopefully) hack, see: https://github.com/jcrist/msgspec/issues/497
            setattr(annotation, "_decoder", decoder)

        if meta_data:
            annotation = Annotated[annotation, meta_data]  # pyright: ignore

        return annotation



from .base_exceptions import LitestarException, LitestarWarning, MissingDependencyException, SerializationException
from .dto_exceptions import DTOFactoryException, InvalidAnnotationException
from .http_exceptions import (
    ClientException,
    HTTPException,
    ImproperlyConfiguredException,
    InternalServerException,
    MethodNotAllowedException,
    NoRouteMatchFoundException,
    NotAuthorizedException,
    NotFoundException,
    PermissionDeniedException,
    ServiceUnavailableException,
    TemplateNotFoundException,
    TooManyRequestsException,
    ValidationException,
)
from .websocket_exceptions import WebSocketDisconnect, WebSocketException

__all__ = (
    "ClientException",
    "DTOFactoryException",
    "HTTPException",
    "ImproperlyConfiguredException",
    "InternalServerException",
    "InvalidAnnotationException",
    "LitestarException",
    "LitestarWarning",
    "MethodNotAllowedException",
    "MissingDependencyException",
    "NoRouteMatchFoundException",
    "NotAuthorizedException",
    "NotFoundException",
    "PermissionDeniedException",
    "SerializationException",
    "ServiceUnavailableException",
    "TemplateNotFoundException",
    "TooManyRequestsException",
    "ValidationException",
    "WebSocketDisconnect",
    "WebSocketException",
)



from __future__ import annotations

from dataclasses import dataclass

from litestar.openapi.spec.base import BaseSchemaObject

__all__ = ("Discriminator",)


@dataclass(unsafe_hash=True)
class Discriminator(BaseSchemaObject):
    """When request bodies or response payloads may be one of a number of different schemas, a ``discriminator``
    object can be used to aid in serialization, deserialization, and validation.

    The discriminator is a specific object in a schema which is used to inform the consumer of the specification of an
    alternative schema based on the value associated with it.

    When using the discriminator, _inline_ schemas will not be considered.
    """

    property_name: str
    """**REQUIRED**. The name of the property in the payload that will hold the discriminator value."""

    mapping: dict[str, str] | None = None
    """An object to hold mappings between payload values and schema names or references."""



from __future__ import annotations

from litestar.testing.client import TestClient


def test_initial_pattern_app():
    from docs.examples.data_transfer_objects.factory.tutorial.initial_pattern import app

    with TestClient(app=app) as client:
        response = client.get("/person/peter")
    assert response.status_code == 200
    assert response.json() == {"name": "peter", "age": 30, "email": "email_of_peter@example.com"}


def test_simple_dto_exclude():
    from docs.examples.data_transfer_objects.factory.tutorial.simple_dto_exclude import app

    with TestClient(app=app) as client:
        response = client.get("/person/peter")
    assert response.status_code == 200
    assert response.json() == {"name": "peter", "age": 30}


def test_nested_exclude():
    from docs.examples.data_transfer_objects.factory.tutorial.nested_exclude import app

    with TestClient(app=app) as client:
        response = client.get("/person/peter")
    assert response.status_code == 200
    assert response.json() == {"name": "peter", "age": 30, "address": {"city": "Cityville", "country": "Countryland"}}


def test_nested_collection_exclude():
    from docs.examples.data_transfer_objects.factory.tutorial.nested_collection_exclude import app

    with TestClient(app=app) as client:
        response = client.get("/person/peter")
    assert response.status_code == 200
    assert response.json() == {
        "name": "peter",
        "age": 30,
        "address": {"city": "Cityville", "country": "Countryland"},
        "children": [{"name": "Child1", "age": 10}, {"name": "Child2", "age": 8}],
    }


def test_max_nested_depth():
    from docs.examples.data_transfer_objects.factory.tutorial.max_nested_depth import app

    with TestClient(app=app) as client:
        response = client.get("/person/peter")
    assert response.status_code == 200
    assert response.json() == {
        "name": "peter",
        "age": 30,
        "address": {"city": "Cityville", "country": "Countryland"},
        "children": [{"name": "Child1", "age": 10, "children": []}, {"name": "Child2", "age": 8, "children": []}],
    }


def test_explicit_field_renaming():
    from docs.examples.data_transfer_objects.factory.tutorial.explicit_field_renaming import app

    with TestClient(app=app) as client:
        response = client.get("/person/peter")
    assert response.status_code == 200
    assert response.json() == {
        "name": "peter",
        "age": 30,
        "location": {"city": "Cityville", "country": "Countryland"},
        "children": [{"name": "Child1", "age": 10}, {"name": "Child2", "age": 8}],
    }


def test_field_renaming_strategy():
    from docs.examples.data_transfer_objects.factory.tutorial.field_renaming_strategy import app

    with TestClient(app=app) as client:
        response = client.get("/person/peter")
    assert response.status_code == 200
    assert response.json() == {
        "NAME": "peter",
        "AGE": 30,
        "ADDRESS": {"CITY": "Cityville", "COUNTRY": "Countryland"},
        "CHILDREN": [{"NAME": "Child1", "AGE": 10}, {"NAME": "Child2", "AGE": 8}],
    }


def test_simple_receiving_data():
    from docs.examples.data_transfer_objects.factory.tutorial.simple_receiving_data import app

    with TestClient(app=app) as client:
        response = client.post("/person", json={"name": "peter", "age": 40, "email": "email_of_peter@example.com"})
    assert response.status_code == 201
    assert response.json() == {"name": "peter", "age": 40}


def test_read_only_fields():
    from docs.examples.data_transfer_objects.factory.tutorial.read_only_fields_error import app

    with TestClient(app=app) as client:
        response = client.post("/person", json={"name": "peter", "age": 40, "email": "email_of_peter@example.com"})

    assert response.status_code == 500


def test_dto_data():
    from docs.examples.data_transfer_objects.factory.tutorial.dto_data import app

    with TestClient(app=app) as client:
        response = client.post("/person", json={"name": "peter", "age": 40, "email": "email_of_peter@example.com"})

    assert response.status_code == 201
    assert response.json() == {"id": 1, "name": "peter", "age": 40}


def test_put_handler():
    from docs.examples.data_transfer_objects.factory.tutorial.put_handlers import app

    with TestClient(app=app) as client:
        response = client.put("/person/1", json={"name": "peter", "age": 50, "email": "email_of_peter@example.com"})

    assert response.status_code == 200
    assert response.json() == {"id": 1, "name": "peter", "age": 50}


def test_patch_handler():
    from docs.examples.data_transfer_objects.factory.tutorial.patch_handlers import app

    with TestClient(app=app) as client:
        response = client.patch("/person/1", json={"name": "peter"})

    assert response.status_code == 200
    assert response.json() == {"id": 1, "name": "peter", "age": 50}


def test_multiple_handlers():
    from docs.examples.data_transfer_objects.factory.tutorial.multiple_handlers import app

    with TestClient(app=app) as client:
        response = client.put("/person/1", json={"name": "peter", "age": 50, "email": "email_of_peter@example.com"})
    assert response.status_code == 200

    with TestClient(app=app) as client:
        response = client.patch("/person/1", json={"name": "peter"})
    assert response.status_code == 200

    with TestClient(app=app) as client:
        response = client.post("/person", json={"name": "peter", "age": 40, "email": "email_of_peter@example.com"})
    assert response.status_code == 201


def test_controller():
    from docs.examples.data_transfer_objects.factory.tutorial.controller import app

    with TestClient(app=app) as client:
        response = client.put("/person/1", json={"name": "peter", "age": 50, "email": "email_of_peter@example.com"})
        assert response.status_code == 200

        response = client.patch("/person/1", json={"name": "peter"})
        assert response.status_code == 200

        response = client.post("/person", json={"name": "peter", "age": 40, "email": "email_of_peter@example.com"})
        assert response.status_code == 201



from __future__ import annotations

from enum import Enum
from functools import partial
from typing import TYPE_CHECKING, Any, TypeVar, cast
from urllib.parse import quote

from litestar.utils.typing import get_origin_or_inner_type

if TYPE_CHECKING:
    from collections.abc import Container

    from litestar.types import MaybePartial

__all__ = (
    "get_enum_string_value",
    "get_name",
    "unwrap_partial",
    "url_quote",
    "unique_name_for_scope",
)

T = TypeVar("T")


def get_name(value: Any) -> str:
    """Get the ``__name__`` of an object.

    Args:
        value: An arbitrary object.

    Returns:
        A name string.
    """

    name = getattr(value, "__name__", None)
    if name is not None:
        return cast("str", name)

    # On Python 3.8 and 3.9, Foo[int] does not have the __name__ attribute.
    if origin := get_origin_or_inner_type(value):
        return cast("str", origin.__name__)

    return type(value).__name__


def get_enum_string_value(value: Enum | str) -> str:
    """Return the string value of a string enum.

    See: https://github.com/litestar-org/litestar/pull/633#issuecomment-1286519267

    Args:
        value: An enum or string.

    Returns:
        A string.
    """
    return value.value if isinstance(value, Enum) else value  # type:ignore


def unwrap_partial(value: MaybePartial[T]) -> T:
    """Unwraps a partial, returning the underlying callable.

    Args:
        value: A partial function.

    Returns:
        Callable
    """
    from litestar.utils.sync import AsyncCallable

    return cast("T", value.func if isinstance(value, (partial, AsyncCallable)) else value)


def url_quote(value: str | bytes) -> str:
    """Quote a URL.

    Args:
        value: A URL.

    Returns:
        A quoted URL.
    """
    return quote(value, safe="/#%[]=:;$&()+,!?*@'~")


def unique_name_for_scope(base_name: str, scope: Container[str]) -> str:
    """Create a name derived from ``base_name`` that's unique within ``scope``"""
    i = 0
    while True:
        if (unique_name := f"{base_name}_{i}") not in scope:
            return unique_name
        i += 1


def get_exception_group() -> type[BaseException]:
    """Get the exception group class with version compatibility."""
    try:
        return cast("type[BaseException]", ExceptionGroup)  # type:ignore[name-defined]
    except NameError:
        from exceptiongroup import ExceptionGroup as _ExceptionGroup

        return cast("type[BaseException]", _ExceptionGroup)



"""Collection filter datastructures."""
from __future__ import annotations

from collections import abc  # noqa: TCH003
from dataclasses import dataclass
from datetime import datetime  # noqa: TCH003
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeVar

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

T = TypeVar("T")

__all__ = (
    "BeforeAfter",
    "CollectionFilter",
    "FilterTypes",
    "LimitOffset",
    "OrderBy",
    "SearchFilter",
    "NotInCollectionFilter",
    "OnBeforeAfter",
    "NotInSearchFilter",
)


FilterTypes: TypeAlias = "BeforeAfter | OnBeforeAfter | CollectionFilter[Any] | LimitOffset | OrderBy | SearchFilter | NotInCollectionFilter[Any] | NotInSearchFilter"
"""Aggregate type alias of the types supported for collection filtering."""


@dataclass
class BeforeAfter:
    """Data required to filter a query on a ``datetime`` column."""

    field_name: str
    """Name of the model attribute to filter on."""
    before: datetime | None
    """Filter results where field earlier than this."""
    after: datetime | None
    """Filter results where field later than this."""


@dataclass
class OnBeforeAfter:
    """Data required to filter a query on a ``datetime`` column."""

    field_name: str
    """Name of the model attribute to filter on."""
    on_or_before: datetime | None
    """Filter results where field is on or earlier than this."""
    on_or_after: datetime | None
    """Filter results where field on or later than this."""


@dataclass
class CollectionFilter(Generic[T]):
    """Data required to construct a ``WHERE ... IN (...)`` clause."""

    field_name: str
    """Name of the model attribute to filter on."""
    values: abc.Collection[T]
    """Values for ``IN`` clause."""


@dataclass
class NotInCollectionFilter(Generic[T]):
    """Data required to construct a ``WHERE ... NOT IN (...)`` clause."""

    field_name: str
    """Name of the model attribute to filter on."""
    values: abc.Collection[T]
    """Values for ``NOT IN`` clause."""


@dataclass
class LimitOffset:
    """Data required to add limit/offset filtering to a query."""

    limit: int
    """Value for ``LIMIT`` clause of query."""
    offset: int
    """Value for ``OFFSET`` clause of query."""


@dataclass
class OrderBy:
    """Data required to construct a ``ORDER BY ...`` clause."""

    field_name: str
    """Name of the model attribute to sort on."""
    sort_order: Literal["asc", "desc"] = "asc"
    """Sort ascending or descending"""


@dataclass
class SearchFilter:
    """Data required to construct a ``WHERE field_name LIKE '%' || :value || '%'`` clause."""

    field_name: str
    """Name of the model attribute to sort on."""
    value: str
    """Values for ``LIKE`` clause."""
    ignore_case: bool | None = False
    """Should the search be case insensitive."""


@dataclass
class NotInSearchFilter:
    """Data required to construct a ``WHERE field_name NOT LIKE '%' || :value || '%'`` clause."""

    field_name: str
    """Name of the model attribute to search on."""
    value: str
    """Values for ``NOT LIKE`` clause."""
    ignore_case: bool | None = False
    """Should the search be case insensitive."""



from typing import List

from litestar import Controller, Router, WebSocket, websocket
from litestar.testing import create_test_client


def test_handle_websocket() -> None:
    @websocket(path="/")
    async def simple_websocket_handler(socket: WebSocket) -> None:
        await socket.accept()
        data = await socket.receive_json()
        assert data
        await socket.send_json({"data": "123"})
        await socket.close()

    client = create_test_client(route_handlers=simple_websocket_handler)

    with client.websocket_connect("/") as ws:
        ws.send_json({"data": "123"})
        data = ws.receive_json()
        assert data


def test_websocket_signature_namespace() -> None:
    class MyController(Controller):
        path = "/ws"
        signature_namespace = {"c": float}

        @websocket(path="/", signature_namespace={"d": List[str]})
        async def simple_websocket_handler(
            self,
            socket: WebSocket,
            a: "a",  # type:ignore[name-defined]  # noqa: F821
            b: "b",  # type:ignore[name-defined]  # noqa: F821
            c: "c",  # type:ignore[name-defined]  # noqa: F821
            d: "d",  # type:ignore[name-defined]  # noqa: F821
        ) -> None:
            await socket.accept()
            data = await socket.receive_json()
            assert data
            await socket.send_json({"a": a, "b": b, "c": c, "d": d})
            await socket.close()

    router = Router("/", route_handlers=[MyController], signature_namespace={"b": str})

    client = create_test_client(route_handlers=[router], signature_namespace={"a": int})

    with client.websocket_connect("/ws?a=1&b=two&c=3.0&d=d") as ws:
        ws.send_json({"data": "123"})
        data = ws.receive_json()
        assert data == {"a": 1, "b": "two", "c": 3.0, "d": ["d"]}



from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Type, Union

import pytest
from jinja2 import DictLoader, Environment
from mako.lookup import TemplateLookup  # type: ignore[import-untyped]

from litestar import get
from litestar.contrib.jinja import JinjaTemplateEngine
from litestar.contrib.mako import MakoTemplateEngine
from litestar.contrib.minijinja import MiniJinjaTemplateEngine
from litestar.response.template import Template
from litestar.template.config import TemplateConfig
from litestar.testing import create_test_client

if TYPE_CHECKING:
    from litestar.handlers.http_handlers import HTTPRouteHandler


@dataclass
class EngineTest:
    engine_class: Optional[Type[Union[JinjaTemplateEngine, MakoTemplateEngine, MiniJinjaTemplateEngine]]]
    index_template: str
    nested_template: str
    instantiated: bool
    instance: Optional[Union[JinjaTemplateEngine, MakoTemplateEngine, MiniJinjaTemplateEngine]]


mako_template_lookup = TemplateLookup()
mako_template_lookup.put_string("index.html", "<html>Injected? ${test}</html>")
mako_template_lookup.put_string("nested-dir/nested.html", "<html>Does nested dirs work? ${test}</html>")
mako_template_lookup.put_string("no_context.html", "<html>This works!</html>")


@pytest.fixture(
    params=[
        EngineTest(
            engine_class=JinjaTemplateEngine,
            index_template="<html>Injected? {{test}}</html>",
            nested_template="<html>Does nested dirs work? {{test}}</html>",
            instantiated=False,
            instance=None,
        ),
        EngineTest(
            engine_class=MakoTemplateEngine,
            index_template="<html>Injected? ${test}</html>",
            nested_template="<html>Does nested dirs work? ${test}</html>",
            instantiated=False,
            instance=None,
        ),
        EngineTest(
            engine_class=MiniJinjaTemplateEngine,
            index_template="<html>Injected? {{test}}</html>",
            nested_template="<html>Does nested dirs work? {{test}}</html>",
            instantiated=False,
            instance=None,
        ),
        EngineTest(
            engine_class=None,
            index_template="<html>Injected? {{test}}</html>",
            nested_template="<html>Does nested dirs work? {{test}}</html>",
            instantiated=True,
            instance=JinjaTemplateEngine.from_environment(
                Environment(
                    loader=DictLoader(
                        {
                            "index.html": "<html>Injected? {{test}}</html>",
                            "nested-dir/nested.html": "<html>Does nested dirs work? {{test}}</html>",
                            "no_context.html": "<html>This works!</html>",
                        }
                    )
                )
            ),
        ),
        EngineTest(
            engine_class=None,
            index_template="<html>Injected? ${test}</html>",
            nested_template="<html>Does nested dirs work? ${test}</html>",
            instantiated=True,
            instance=MakoTemplateEngine.from_template_lookup(mako_template_lookup),
        ),
    ]
)
def engine_test(request: Any) -> EngineTest:
    return request.param  # type:ignore[no-any-return]


@pytest.fixture()
def index_handler(engine_test: EngineTest, tmp_path: Path) -> "HTTPRouteHandler":
    Path(tmp_path / "index.html").write_text(engine_test.index_template)

    @get(path="/")
    def index_handler() -> Template:
        return Template(template_name="index.html", context={"test": "yep"})

    return index_handler


@pytest.fixture()
def nested_path_handler(engine_test: EngineTest, tmp_path: Path) -> "HTTPRouteHandler":
    nested_path = tmp_path / "nested-dir"
    nested_path.mkdir()
    Path(nested_path / "nested.html").write_text(engine_test.nested_template)

    @get(path="/nested")
    def nested_path_handler() -> Template:
        return Template(template_name="nested-dir/nested.html", context={"test": "yep"})

    return nested_path_handler


@pytest.fixture()
def template_config(engine_test: EngineTest, tmp_path: Path) -> TemplateConfig:
    if engine_test.instantiated:
        return TemplateConfig(instance=engine_test.instance)
    return TemplateConfig(engine=engine_test.engine_class, directory=tmp_path)


def test_template(index_handler: "HTTPRouteHandler", template_config: TemplateConfig) -> None:
    with create_test_client(route_handlers=[index_handler], template_config=template_config) as client:
        response = client.request("GET", "/")
        assert response.status_code == 200, response.text
        assert response.text == "<html>Injected? yep</html>"
        assert response.headers["Content-Type"] == "text/html; charset=utf-8"


def test_nested_tmp_pathectory(nested_path_handler: "HTTPRouteHandler", template_config: TemplateConfig) -> None:
    with create_test_client(route_handlers=[nested_path_handler], template_config=template_config) as client:
        response = client.request("GET", "/nested")
        assert response.status_code == 200, response.text
        assert response.text == "<html>Does nested dirs work? yep</html>"
        assert response.headers["Content-Type"] == "text/html; charset=utf-8"


def test_raise_for_invalid_template_name(template_config: TemplateConfig) -> None:
    @get(path="/")
    def invalid_template_name_handler() -> Template:
        return Template(template_name="invalid.html", context={"test": "yep"})

    with create_test_client(
        route_handlers=[invalid_template_name_handler], template_config=template_config, debug=False
    ) as client:
        response = client.request("GET", "/")
        assert response.status_code == 500
        assert response.json() == {"detail": "Internal Server Error", "status_code": 500}


def test_no_context(tmp_path: Path, template_config: TemplateConfig) -> None:
    Path(tmp_path / "no_context.html").write_text("<html>This works!</html>")

    @get(path="/")
    def index() -> Template:
        return Template(template_name="no_context.html")

    with create_test_client(route_handlers=[index], template_config=template_config) as client:
        index_response = client.request("GET", "/")
        assert index_response.status_code == 200
        assert index_response.text == "<html>This works!</html>"
        assert index_response.headers["Content-Type"] == "text/html; charset=utf-8"



from __future__ import annotations

from dataclasses import asdict, dataclass, fields, is_dataclass
from enum import Enum
from typing import Any

__all__ = ("BaseSchemaObject",)


def _normalize_key(key: str) -> str:
    if key.endswith("_in"):
        return "in"
    if key.startswith("schema_"):
        return key.split("_")[1]
    if "_" in key:
        components = key.split("_")
        return components[0] + "".join(component.title() for component in components[1:])
    return "$ref" if key == "ref" else key


def _normalize_value(value: Any) -> Any:
    if isinstance(value, BaseSchemaObject):
        return value.to_schema()
    if is_dataclass(value):
        return {_normalize_value(k): _normalize_value(v) for k, v in asdict(value).items() if v is not None}
    if isinstance(value, dict):
        return {_normalize_value(k): _normalize_value(v) for k, v in value.items() if v is not None}
    if isinstance(value, list):
        return [_normalize_value(v) for v in value]
    return value.value if isinstance(value, Enum) else value


@dataclass
class BaseSchemaObject:
    """Base class for schema spec objects"""

    def to_schema(self) -> dict[str, Any]:
        """Transform the spec dataclass object into a string keyed dictionary. This method traverses all nested values
        recursively.
        """
        result: dict[str, Any] = {}

        for field in fields(self):
            value = _normalize_value(getattr(self, field.name, None))

            if value is not None:
                if "alias" in field.metadata:
                    if not isinstance(field.metadata["alias"], str):
                        raise TypeError('metadata["alias"] must be a str')
                    key = field.metadata["alias"]
                else:
                    key = _normalize_key(field.name)

                result[key] = value

        return result
